#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcases:
    env_flag = sys.argv[1]
    test = test_gbp_ptg_func(env_flag)
    test.global_cfg()
    if test.test_gbp_ptg_func_1() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PTG_FUNC_1')
        test.global_cfg()  # Making global_cfg available for the subsequent TC
    if test.test_gbp_ptg_func_2() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PTG_FUNC_2')
        test.global_cfg()
    if test.test_gbp_ptg_func_3() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PTG_FUNC_3')
    test.cleanup()
    utils_libs.report_results('test_gbp_ptg_func', 'test_results.txt')
    sys.exit(1)


class test_gbp_ptg_func(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    hdlr = logging.FileHandler('/tmp/test_gbp_ptg_func.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self, env_flag):
        """
        Init def
        """
        self._log.info(
            "\n## START OF GBP POLICY_TARGET_GROUP FUNCTIONALITY TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.act_name = 'test_ptg_pa'
        self.cls_name = 'test_ptg_pc'
        self.rule_name = 'test_ptg_pr'
        self.ruleset_name = 'test_ptg_prs'
        self.ptg_name = 'demo_ptg'
        self.l2p_name = 'test_ptg_l2p'
        self.l3p_name = 'test_ptg_l3p'
        self.pt_name = 'test_pt'
        self.env_flag = env_flag
        if self.env_flag == 'aci':
            self.def_ip_pool = '192.168.0.0/16'
            self.cidr = '192.168.0.0/24'
        else:
            self.def_ip_pool = '10.0.0.0/8'
            self.cidr = '10.0.0.0/24'

    def global_cfg(self):
        self._log.info('\n## Step 1: Create a PC needed for PTG Testing ##')
        self.cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', self.cls_name)
        if self.cls_uuid == 0:
            self._log.info(
                "\nReqd Policy Classifier Create Failed, hence GBP "
                "Policy Target-Group Functional Test Suite Run ABORTED\n")
            return 0
        self._log.info('\n## Step 1: Create a PA needed for PTG Testing ##')
        self.act_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'action', self.act_name)
        if self.act_uuid == 0:
            self._log.info(
                "\n## Reqd Policy Action Create Failed, hence GBP "
                "Policy Target-Group Functional Test Suite Run ABORTED\n")
            return 0
        self._log.info('\n## Step 1: Create a PR needed for PTG Testing ##')
        self.rule_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'rule', self.rule_name, classifier=self.cls_name,
            action=self.act_name)
        if self.rule_uuid == 0:
            self._log.info(
                "\n## Reqd Policy Rule Create Failed, hence GBP Policy "
                "Target-Group Functional Test Suite Run ABORTED\n ")
            return 0
        self._log.info('\n## Step 1: Create a PRS needed for PTG Testing ##')
        self.prs_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'ruleset', self.ruleset_name, policy_rules=self.rule_name)
        if self.prs_uuid == 0:
            self._log.info(
                "\n## Reqd Policy Target-Group Create Failed, hence "
                "GBP Policy Target-Group Functional Test Suite "
                "Run ABORTED\n ")
            return 0
        l3p_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'l3p', self.l3p_name, ip_pool='20.20.0.0/24',
            subnet_prefix_length='28', _proxy_ip_pool='20.20.1.0/24',
            _proxy_subnet_prefix_length='28')
        if l3p_uuid == 0:
            self._log.info(
                "\n## Reqd L3Policy Create Failed, hence GBP Policy "
                "Target-Group Functional Test Suite Run ABORTED\n")
            return 0
        self.gbpcfg.gbp_policy_cfg_all(
            1, 'l2p', self.l2p_name, l3_policy=l3p_uuid)

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('%s: FAILED' % (tc_name))
        for obj in [
                'target',
                'group',
                'l2p',
                'l3p',
                'ruleset',
                'rule',
                'classifier',
                'action']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_gbp_ptg_func_1(
            self,
            name_uuid='',
            ptg_uuid='',
            rep_cr=0,
            rep_del=0):

        if rep_cr == 0 and rep_del == 0:
            self._log.info(
                "\n########################################################\n"
                "TESTCASE_GBP_PTG_FUNC_1: TO CREATE/VERIFY/DELETE/VERIFY a "
                "POLICY TARGET-GROUP with DEFAULT ATTRIB VALUE\n"
                "TEST_STEP::\n"
                "Create Policy Target-Group Object\n"
                "Verify the attributes & value, show & list cmds\n"
                "Verify the implicitly GBP(L2P,L3P) & "
                "Neutron(net,subnet,dhcp-port) Objects\n"
                "Delete Policy Target-Group using Name\n"
                "Verify the PTG has got deleted, show & list cmds\n"
                "Verify the implicit GBP & Neutron Objects are deleted\n"
                "##########################################################\n")

        if name_uuid == '':
            name_uuid = self.ptg_name
        # Testcase work-flow starts
        if rep_cr == 0 or rep_cr == 1:
            self._log.info(
                '\n## Step 1: Create Target-Group with default '
                'attrib vals##\n')
            uuids = self.gbpcfg.gbp_policy_cfg_all(1, 'group', name_uuid)
            if uuids != 0:
                ptg_uuid = uuids[0]
                l2pid = uuids[1]
                subnetid = uuids[2]
            else:
                self._log.info("\n## Step 1: Create Target-Group == Failed")
                return 0
            self._log.info('\n## Step 2A: Verify Target-Group using -list cmd')
            if self.gbpverify.gbp_policy_verify_all(
                    0, 'group', name_uuid, ptg_uuid) == 0:
                self._log.info(
                    "\n## Step 2A: Verify Target-Group using -list "
                    "option == Failed")
                return 0
            self._log.info('\n## Step 2B: Verify Target-Group using -show cmd')
            if self.gbpverify.gbp_policy_verify_all(
                    1, 'group', name_uuid, id=ptg_uuid, shared='False') == 0:
                self._log.info(
                    "\n## Step 2B: Verify Target-Group using -show "
                    "option == Failed")
                return 0
            # Verify the implicit objects(gbp & neutron)
            ret_uuid = self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1, 'l2p', self.ptg_name, ret='default', id=l2pid,
                policy_target_groups=ptg_uuid)
            if ret_uuid != 0 and len(ret_uuid) == 2:
                l3pid = ret_uuid[0]
                ntkid = ret_uuid[1]
            else:
                self._log.info(
                    "\n## Step 2C: Verify By-Default L2Policy == Failed")
                return 0
            rtr_uuid = self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1,
                'l3p',
                l3pid,
                ret='default',
                id=l3pid,
                name='default',
                ip_pool=self.def_ip_pool,
                l2_policies=l2pid,
                subnet_prefix_length='24',
                ip_version='4')
            if rtr_uuid != 0 and isinstance(rtr_uuid, str) == 0:
                self._log.info(
                    "# Step 2D: Verify By-Default L3Policy == Failed")
                return 0
            net_name = 'l2p_%s' % (name_uuid)
            if self.gbpverify.neut_ver_all(
                    'net',
                    ntkid,
                    name=net_name,
                    admin_state_up='True',
                    subnets=subnetid) == 0:
                self._log.info(
                    "# Step 2E: Implicit-creation of Neutron Network-Obj "
                    "-show option == Failed")
                return 0
            if self.gbpverify.neut_ver_all(
                    'subnet',
                    subnetid,
                    cidr=self.cidr,
                    enable_dhcp='True',
                    network_id=ntkid) == 0:
                self._log.info(
                    "\n## Step 2F: Implicit-creation of Neutron SubNet-Obj "
                    "== Failed")
                return 0
            if self.env_flag != 'aci':
                if self.gbpverify.neut_ver_all(
                        'router',
                        rtr_uuid,
                        admin_state_up='True',
                        status='ACTIVE') == 0:
                    self._log.info(
                        "\n## Step 2G: Implicit-creation of Neutron "
                        "Router-Obj == Failed")
                    return 0
        # Delete and Verify
        if rep_del == 0 or rep_del > 0:
            self._log.info('\n## Step 3: Delete Target-Group using name  ##\n')
            if self.gbpcfg.gbp_policy_cfg_all(0, 'group', ptg_uuid) == 0:
                self._log.info("\n## Step 3: Delete Target-Group == Failed")
                return 0
            if self.gbpverify.gbp_policy_verify_all(0, 'group', ptg_uuid) != 0:
                self._log.info(
                    "\n## Step 3A: Verify Target-Group is Deleted "
                    "using -list option == Failed")
                return 0
            if self.gbpverify.gbp_policy_verify_all(1, 'group', ptg_uuid) != 0:
                self._log.info(
                    "\n## Step 3B: Verify Target-Group is Deleted "
                    "using -show option == Failed")
                return 0
            if rep_cr == 0 and rep_del == 0:
                self._log.info("\n## TESTCASE_GBP_PTG_FUNC_1: PASSED")
        return 1

    def test_gbp_ptg_func_2(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PTG_FUNC_2: TO CREATE/VERIFY/DELETE/VERIFY "
            "a POLICY TARGET-GROUP with POLICY RULESET\n"
            "TEST_STEPS::\n"
            "Create Policy Target-Group Object with ConsumedPRS=A\n"
            "Verify the attributes & value, show & list cmds\n"
            "Update the PTG's atribute ProvidedPRS=A\n"
            "Create a PRS=B\n"
            "Update the PTG's attributes Consumed & Provided PRS=B\n"
            "Delete Policy Target-Group using Name\n"
            "Verify that Target-Group has got deleted, show & list cmds\n"
            "##############################################################\n")

        # Testcase work-flow starts
        self._log.info("\n## Step 1: Create Policy Target-Group with PRS ##")
        uuids = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'group',
            self.ptg_name,
            consumed_policy_rule_sets='%s=scope' %
            (self.ruleset_name))
        if uuids != 0:
            ptg_uuid = uuids[0].rstrip()
            subnetid = uuids[2].rstrip()
        else:
            self._log.info("\n## Step 1: Create Target-Group == Failed")
            return 0
        self._log.info(
            '\n## Step 2A: Verify Policy Target-Group using -list cmd')
        if self.gbpverify.gbp_policy_verify_all(
                0, 'group', self.ptg_name, ptg_uuid) == 0:
            self._log.info(
                "\n## Step 2A: Verify Target-Group using -list "
                "option == Failed")
            return 0
        self._log.info(
            '\n## Step 2B: Verify Policy Target-Group using -show cmd')
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'group',
                self.ptg_name,
                id=ptg_uuid,
                shared='False',
                subnets=subnetid,
                consumed_policy_rule_sets=self.prs_uuid) == 0:
            self._log.info(
                "\n## Step 2B: Verify Policy Target-Group using -show "
                "option == Failed")
            return 0
        # Update the PTG's Provided PRS
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'group', ptg_uuid, provided_policy_rule_sets='%s=scope' %
                (self.ruleset_name), name='ptg_new') == 0:
            self._log.info(
                "\n## Step 3: Updating Policy Target-Group == Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'group',
                'ptg_new',
                id=ptg_uuid,
                shared='False',
                subnets=subnetid,
                consumed_policy_rule_sets=self.prs_uuid,
                provided_policy_rule_sets=self.prs_uuid) == 0:
            self._log.info(
                "\n## Step 3A: Verify after updating Policy "
                "Target-Group == Failed")
            return 0
        # Create new PRS and update both Provided & Consumed PRS attrs
        new_prs_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'ruleset', 'demo-new-prs', policy_rules=self.rule_name)
        if new_prs_uuid == 0:
            self._log.info(
                "\n## Step 4: Reqd Policy Target-Group Create Failed, "
                "hence Testcase_gbp_ptg_func_2 Run ABORTED\n ")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(
                2,
                'group',
                ptg_uuid,
                provided_policy_rule_sets='demo-new-prs=scope',
                consumed_policy_rule_sets='demo-new-prs=scope') == 0:
            self._log.info(
                "\n## Step 5: Updating Policy Target-Group with new "
                "PRS == Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'group',
                'ptg_new',
                id=ptg_uuid,
                shared='False',
                subnets=subnetid,
                consumed_policy_rule_sets=new_prs_uuid,
                provided_policy_rule_sets=new_prs_uuid) == 0:
            self._log.info(
                "\n## Step 5A: Verify after updating Policy "
                "Target-Group == Failed")
            return 0
        # Delete the PTG and verify
        self.test_gbp_ptg_func_1(ptg_uuid=ptg_uuid, rep_del=2, rep_cr=2)
        self._log.info("\n## TESTCASE_GBP_PTG_FUNC_2: PASSED")
        return 1

    def test_gbp_ptg_func_3(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PTG_FUNC_3: TO UPDATE A POLICY "
            "TARGET-GROUP AFTER DELETING PT's NEUTRON PORT \n"
            "TEST_STEPS::\n"
            "Create Policy Target-Group using L2P and NO PRS\n"
            "Create a Policy Target using the above Policy-Target-Group\n"
            "Delete the neutron port corresponding to the Policy-Target\n"
            "Update the Policy-Target-Group with a PRS\n"
            "Verify Policy Target-Group successfully updated\n"
            "##############################################################\n")

        # Testcase work-flow starts
        self._log.info('\n## Step 1: Create Policy Target-Group with L2P ##\n')
        uuids = self.gbpcfg.gbp_policy_cfg_all(
            1, 'group', self.ptg_name, l2_policy=self.l2p_name)
        if uuids != 0:
            ptg_uuid = uuids[0]
        else:
            self._log.info("\n## Step 1: Create Target-Group == Failed")
            return 0
        self._log.info(
            '\n## Step 2: Create a Policy Target using the above '
            'Policy-Target-Group\n')
        uuids = self.gbpcfg.gbp_policy_cfg_all(
            1, 'target', self.pt_name, policy_target_group=ptg_uuid)
        if uuids != 0:
            pt_uuid = uuids[0]
            neutron_port_id = uuids[1]
        else:
            self._log.info("\n## Step 2: Create Policy Target == Failed")
            return 0
        self._log.info(
            '\n## Step 2A: Verify the Implicit creation of Neutron Port\n')
        if self.gbpverify.neut_ver_all('port', neutron_port_id) == 0:
            self._log.info(
                "\n## Step 2A: Implicit creation neutron port-object "
                "== Failed")
            return 0
        self._log.info(
            '\n## Step 3: Delete the neutron port corresponding to the '
            'Policy-Target\n')
        cmd = 'neutron port-delete %s' % (neutron_port_id)
        if self.gbpcfg.cmd_error_check(commands.getoutput(cmd)) == 0:
            self._log.info(
                "\n## Step 3: Deletion of the neutron port corresponding "
                "to the Policy-Target = Failed")
            return 0
        self._log.info(
            '\n## Step 4: Update the Policy-Target-Group with a PRS\n')
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'group', ptg_uuid, provided_policy_rule_sets='%s=scope' %
                (self.prs_uuid), consumed_policy_rule_sets='%s=scope' %
                (self.prs_uuid)) == 0:
            self._log.info(
                "\n## Step 4: Updating Policy Target-Group with "
                "new PRS == Failed")
            return 0
        self._log.info(
            '\n## Step 5: Verify Policy Target-Group successfully updated\n')
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'group',
                self.ptg_name,
                id=ptg_uuid,
                shared='False',
                policy_targets=pt_uuid,
                consumed_policy_rule_sets=self.prs_uuid,
                provided_policy_rule_sets=self.prs_uuid) == 0:
            self._log.info(
                "\n## Step 5A: Verify after updating Policy "
                "Target-Group == Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_PTG_FUNC_3: PASSED")
        return 1


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneclient import exceptions as k_exceptions
from keystoneclient.v2_0 import client as keyclient

from gbpservice.common import utils
from gbpservice.contrib.nfp.config_orchestrator.common import topics
from gbpservice.nfp.core import log as nfp_logging
import netaddr

from neutron.db import l3_db
from neutron.db.l3_db import DEVICE_OWNER_ROUTER_INTF
from neutron.db.l3_db import EXTERNAL_GW_INFO
from neutron.db.models.l3 import RouterPort
from neutron.db import models_v2
from neutron_lib import constants as nlib_const
from neutron_lib import exceptions as n_exc
from neutron_lib.exceptions import l3
from neutron_lib.plugins import constants as n_const

import neutron_fwaas.extensions
from neutron_fwaas.services.firewall import fwaas_plugin as ref_fw_plugin
from oslo_config import cfg
from oslo_utils import excutils
from oslo_utils import uuidutils
from sqlalchemy import orm

from neutron_fwaas.db.firewall import firewall_db as n_firewall

LOG = nfp_logging.getLogger(__name__)


class NFPFirewallPlugin(ref_fw_plugin.FirewallPlugin):
    def __init__(self):
        # Monkey patch L3 agent topic
        # L3 agent was where reference firewall agent runs
        # patch that topic to the NFP firewall agent's topic name
        ref_fw_plugin.f_const.L3_AGENT = topics.FW_NFP_CONFIGAGENT_TOPIC

        # Ensure neutron fwaas extensions are loaded
        ext_path = neutron_fwaas.extensions.__path__[0]
        if ext_path not in cfg.CONF.api_extensions_path.split(':'):
            cfg.CONF.set_override(
                'api_extensions_path',
                cfg.CONF.api_extensions_path + ':' + ext_path)

        super(NFPFirewallPlugin, self).__init__()

    # Modifying following plugin function, to relax same router validation
    def _get_routers_for_create_firewall(self, tenant_id, context, firewall):

        # pop router_id as this goes in the router association db
        # and not firewall db
        router_ids = firewall['firewall'].pop('router_ids', None)
        if router_ids == nlib_const.ATTR_NOT_SPECIFIED:
            return tenant_id

    def set_routers_for_firewall(self, context, fw):
        """Sets the routers associated with the fw."""
        pass

    def get_firewall_routers(self, context, fwid):
        """Gets all routers associated with a firewall."""
        fw_rtrs = ['1234567890']
        return fw_rtrs

    def validate_firewall_routers_not_in_use(
            self, context, router_ids, fwid=None):
        """Validate if router-ids not associated with any firewall.

        If any of the router-ids in the list is already associated with
        a firewall, raise an exception else just return.
        """
        pass

    def update_firewall_routers(self, context, fw):
        """Update the firewall with new routers.

        This involves removing existing router associations and replacing
        it with the new router associations provided in the update method.
        """
        return fw


# Monkey patching the create_firewall db method
def create_firewall(self, context, firewall, status=None):
    fw = firewall['firewall']
    tenant_id = fw['tenant_id']
    # distributed routers may required a more complex state machine;
    # the introduction of a new 'CREATED' state allows this, whilst
    # keeping a backward compatible behavior of the logical resource.
    if not status:
        status = n_const.PENDING_CREATE
    with context.session.begin(subtransactions=True):
        self._validate_fw_parameters(context, fw, tenant_id)
        firewall_db = n_firewall.Firewall(
                id=uuidutils.generate_uuid(),
                tenant_id=tenant_id,
                name=fw['name'],
                description=fw['description'],
                firewall_policy_id=fw['firewall_policy_id'],
                admin_state_up=fw['admin_state_up'],
                status=status)
        context.session.add(firewall_db)
    return self._make_firewall_dict(firewall_db)


n_firewall.Firewall_db_mixin.create_firewall = create_firewall


# Monkey patching l3_db's _get_router_for_floatingip method to associate
# floatingip if corresponding routes is present.
def _is_net_reachable_from_net(self, context, tenant_id, from_net_id,
                               to_net_id):
    """Check whether a network is reachable.

    Follow the paths of networks connected by devices, to determine
    whether a network is reachable from another.
    @param context: neutron api request context
    @param tenant_id: the owning tenant
    @param from_net_id: the source network for the search
    @param to_net_id: the destination network for the search
    @return: True or False whether a path exists
    """
    original_context = context
    context = elevate_context(context)
    tenant_id = context.tenant_id

    def nexthop_nets_query(nets, visited):
        """query networks connected to devices on nets but not visited."""
        Port = models_v2.Port
        devices_on_nets = context.session.query(Port.device_id).filter(
            Port.tenant_id == tenant_id,
            Port.device_owner.notin_([nlib_const.DEVICE_OWNER_DHCP]),
            Port.network_id.in_(nets)).subquery()
        return context.session.query(Port.network_id).filter(
            Port.tenant_id == tenant_id,
            Port.network_id.notin_(visited),
            Port.device_id.in_(devices_on_nets))
    visited = set([])
    nets = set([from_net_id])
    while nets:
        if to_net_id in nets:
            context = original_context
            return True
        visited |= nets
        nets = set((tup[0] for tup in nexthop_nets_query(nets, visited)))
    context = original_context
    return False


def _find_net_for_nexthop(self, context, tenant_id, router_id, nexthop):
    """Find the network to which the nexthop belongs.

    Iterate over the router interfaces to find the network of nexthop.
    @param context: neutron api request context
    @param tenant_id: the owning tenant
    @param router_id: a router id
    @param nexthop: an IP address
    @return: the network id of the nexthop or None if not found
    """
    interfaces = context.session.query(models_v2.Port).filter_by(
        tenant_id=tenant_id,
        device_id=router_id,
        device_owner=DEVICE_OWNER_ROUTER_INTF)
    for interface in interfaces:
        cidrs = [self._core_plugin._get_subnet(context,
                                               ip['subnet_id'])['cidr']
                 for ip in interface['fixed_ips']]
        if netaddr.all_matching_cidrs(nexthop, cidrs):
            return interface['network_id']


def _find_routers_via_routes_for_floatingip(self, context, internal_port,
                                            internal_subnet_id,
                                            external_network_id):
    """Find routers with route to the internal IP address.

    Iterate over the routers that belong to the same tenant as
    'internal_port'. For each router check that the router is connected
    to the external network and whether there is a route to the internal
    IP address. Consider only routers for which there is a path from the
    nexthop of the route to the internal port.

    Sort the list of routers to have the router with the most specific
    route first (largest CIDR prefix mask length).

    @param context: neutron api request context
    @param internal_port: the port dict for the association
    @param internal_subnet_id: the subnet for the association
    @param external_network_id: the network of the floatingip
    @return: a sorted list of matching routers
    """
    original_context = context
    context = elevate_context(context)
    internal_ip_address = [
        ip['ip_address'] for ip in internal_port['fixed_ips']
        if ip['subnet_id'] == internal_subnet_id
    ][0]

    # find the tenant routers
    tenant_id = internal_port['tenant_id']
    routers = self.get_routers(context, filters={'tenant_id': [tenant_id]})

    prefix_routers = []
    for router in routers:
        # verify that the router is on "external_network"
        gw_info = router.get(EXTERNAL_GW_INFO)
        if not gw_info or gw_info['network_id'] != external_network_id:
            continue
        # find a matching route
        if 'routes' not in router:
            continue
        cidr_nexthops = {}
        for route in router['routes']:
            cidr = netaddr.IPNetwork(route['destination'])
            if cidr not in cidr_nexthops:
                cidr_nexthops[cidr] = []
            cidr_nexthops[cidr].append(route['nexthop'])
        smallest_cidr = netaddr.smallest_matching_cidr(
            internal_ip_address,
            cidr_nexthops.keys())
        if not smallest_cidr:
            continue
        # validate that there exists a path to "internal_port"
        for nexthop in cidr_nexthops[smallest_cidr]:
            net_id = self._find_net_for_nexthop(context, context.tenant_id,
                                                router['id'], nexthop)
            if net_id and self._is_net_reachable_from_net(
                    context,
                    context.tenant_id,
                    net_id,
                    internal_port['network_id']):
                prefix_routers.append(
                    (smallest_cidr.prefixlen, router['id']))
                break
    context = original_context
    return [p_r[1] for p_r in sorted(prefix_routers, reverse=True)]


def elevate_context(context):
    context = context.elevated()
    context.tenant_id = _resource_owner_tenant_id()
    return context


def _resource_owner_tenant_id():
    user, pwd, tenant, auth_url = utils.get_keystone_creds()
    keystoneclient = keyclient.Client(username=user, password=pwd,
                                      auth_url=auth_url)
    try:
        tenant = keystoneclient.tenants.find(name=tenant)
        return tenant.id
    except k_exceptions.NotFound:
        with excutils.save_and_reraise_exception(reraise=True):
            LOG.error('No tenant with name %s exists.', tenant)
    except k_exceptions.NoUniqueMatch:
        with excutils.save_and_reraise_exception(reraise=True):
            LOG.error('Multiple tenants matches found for %s', tenant)


def _get_router_for_floatingip(self, context, internal_port,
                               internal_subnet_id,
                               external_network_id):
    subnet = self._core_plugin.get_subnet(context, internal_subnet_id)
    if not subnet['gateway_ip']:
        msg = (_('Cannot add floating IP to port on subnet %s '
                 'which has no gateway_ip') % internal_subnet_id)
        raise n_exc.BadRequest(resource='floatingip', msg=msg)

    # Find routers(with router_id and interface address) that
    # connect given internal subnet and the external network.
    # Among them, if the router's interface address matches
    # with subnet's gateway-ip, return that router.
    # Otherwise return the first router.
    gw_port = orm.aliased(models_v2.Port, name="gw_port")
    routerport_qry = context.session.query(
        RouterPort.router_id, models_v2.IPAllocation.ip_address).join(
        models_v2.Port, models_v2.IPAllocation).filter(
        models_v2.Port.network_id == internal_port['network_id'],
        RouterPort.port_type.in_(nlib_const.ROUTER_INTERFACE_OWNERS),
        models_v2.IPAllocation.subnet_id == internal_subnet_id
    ).join(gw_port, gw_port.device_id == RouterPort.router_id).filter(
        gw_port.network_id == external_network_id).distinct()

    first_router_id = None
    for router_id, interface_ip in routerport_qry:
        if interface_ip == subnet['gateway_ip']:
            return router_id
        if not first_router_id:
            first_router_id = router_id
    if first_router_id:
        return first_router_id

    router_ids = self._find_routers_via_routes_for_floatingip(
        context,
        internal_port,
        internal_subnet_id,
        external_network_id)
    if router_ids:
        return router_ids[0]

    raise l3.ExternalGatewayForFloatingIPNotFound(
        subnet_id=internal_subnet_id,
        external_network_id=external_network_id,
        port_id=internal_port['id'])


l3_db.L3_NAT_dbonly_mixin._get_router_for_floatingip = (
        _get_router_for_floatingip)
l3_db.L3_NAT_dbonly_mixin._find_routers_via_routes_for_floatingip = (
        _find_routers_via_routes_for_floatingip)
l3_db.L3_NAT_dbonly_mixin._find_net_for_nexthop = _find_net_for_nexthop
l3_db.L3_NAT_dbonly_mixin._is_net_reachable_from_net = (
        _is_net_reachable_from_net)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcases:
    test = test_gbp_nsp_func()
    if test.test_gbp_nsp_func_1() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_NSP_FUNC_1')
    if test.test_gbp_nsp_func_2() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_NSP_FUNC_2')
    if test.test_gbp_nsp_func_3() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_NSP_FUNC_3')
    test.cleanup()
    utils_libs.report_results('test_gbp_nsp_func', 'test_results.txt')
    sys.exit(1)


class test_gbp_nsp_func(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_nsp_func.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_nsp_func.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info(
            "\n## START OF GBP NETWORK_SERVICE_POLICY FUNCTIONALITY "
            "TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.nsp_name = 'demo_nsp'

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('%s: FAILED' % (tc_name))
        for obj in ['group', 'nsp']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_gbp_nsp_func_1(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_NSP_FUNC_1: TO CREATE/REFER/DELETE/VERIFY "
            "NTK-SVC-POLICY in PTG\n"
            "TEST_STEPS::\n"
            "Create two NSPs one with type:ip-pool & ip-single, "
            "value:self_subnet and self_subnet\n"
            "Verify the attributes & values\n"
            "Create two PTGs and reference each one of the above "
            "NSP in one of the PTG\n"
            "Verify the NSP reference in the PTGs\n"
            "Delete the PTG and the NSP\n"
            "Verify that NSP got deleted\n"
            "##############################################################\n")

        # Testcase work-flow starts
        # Create and Verify NSPolicy with type=ip_single & ip-single,
        # name:self_subnet & self_subnet
        self._log.info(
            '\n## Step 1: Create NSPolicy with type=ip_single & '
            'name:self_subnet ##\n')
        nsp1_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'nsp',
            'demo_nsp_1',
            network_service_params="type=ip_single,name=vip_ip1,"
                                   "value=self_subnet")
        if nsp1_uuid == 0:
            self._log.info(
                "\n## Step 1A: Create NSPolicy with type=ip_single & "
                "name:self_subnet == Failed")
            return 0
        nsp2_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'nsp',
            'demo_nsp_2',
            network_service_params="type=ip_single,name=vip_ip2,"
                                   "value=self_subnet")
        if nsp2_uuid == 0:
            self._log.info(
                "\n## Step 1B: Create NSPolicy with type=ip_single & "
                "name:self_subnet == Failed")
            return 0
        # Verify
        self._log.info(
            "\n## Step 2: Verify NSPolicies are successfully created")
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1,
                'nsp',
                nsp1_uuid,
                name='demo_nsp_1',
                network_service_params='{"type": "ip_single", "name": '
                                       '"vip_ip1", "value": '
                                       '"self_subnet"}') == 0:
            self._log.info(
                "\n## Step 2A: Verify NSPolicy demo_nsp_1 with valued "
                "attributes, Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1,
                'nsp',
                nsp2_uuid,
                name='demo_nsp_2',
                network_service_params='{"type": "ip_single", '
                                       '"name": "vip_ip2", "value": '
                                       '"self_subnet"}') == 0:
            self._log.info(
                "\n## Step 2A: Verify NSPolicy demo_nsp_2 with "
                "valued attributes, Failed")
            return 0
        # Create two PTGs, each referencing one of the two NSPs
        self._log.info(
            "\n## Step 3: Create and Verify two PTGs each "
            "referencing one of the two NSPs")
        uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'group', 'demo_ptg_1', network_service_policy=nsp1_uuid)
        if uuid == 0:
            self._log.info(
                "\n## Step 3A: Create PTG using NSP demo_nsp_1,Failed")
            return 0
        else:
            ptg1_uuid = uuid[0]
        _uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'group', 'demo_ptg_2', network_service_policy=nsp2_uuid)
        if _uuid == 0:
            self._log.info(
                "\n## Step 3B: Create PTG using NSP demo_nsp_2,Failed")
            return 0
        else:
            ptg2_uuid = _uuid[0]
        # Verify
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1, 'nsp', nsp1_uuid, policy_target_groups=ptg1_uuid) == 0:
            self._log.info(
                "\n## Step 3C: Verify PTG demo_ptg_1 seen in NSP "
                "demo_nsp_1, Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1, 'nsp', nsp2_uuid, policy_target_groups=ptg2_uuid) == 0:
            self._log.info(
                "\n## Step 3C: Verify PTG demo_ptg_2 seen in NSP "
                "demo_nsp_2, Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1, 'group', ptg1_uuid,
                network_service_policy_id=nsp1_uuid) == 0:
            self._log.info(
                "\n## Step 3D: Verify PTG demo_ptg_1 references NSP "
                "demo_nsp_1, Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1, 'group', ptg2_uuid,
                network_service_policy_id=nsp2_uuid) == 0:
            self._log.info(
                "\n## Step 3D: Verify PTG demo_ptg_2 references NSP "
                "demo_nsp_2, Failed")
            return 0
        # Delete PTGs & NSPs
        self._log.info(
            "\n## Step 4: Delete and Verify two PTGs each referencing "
            "one of the two NSPs")
        ptg_list = [ptg1_uuid, ptg2_uuid]
        nsp_list = [nsp1_uuid, nsp2_uuid]
        for i in range(len(ptg_list)):
            if self.gbpcfg.gbp_policy_cfg_all(0, 'group', ptg_list[i]) == 0:
                self._log.info(
                    "\n## Step 4A: Deletion of PTG %s, Failed" %
                    (ptg_list[i]))
                return 0
            if self.gbpcfg.gbp_policy_cfg_all(0, 'nsp', nsp_list[i]) == 0:
                self._log.info(
                    "\n## Step 4B: Deletion of NSP %s, Failed" %
                    (nsp_list[i]))
                return 0
        # Verify
        for n in range(len(nsp_list)):
            if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                    1, 'nsp', nsp_list[n]) != 0:
                self._log.info("\n## Step 4C: Verify deletion of NSP, Failed")
                return 0
        self._log.info("\n## TESTCASE_GBP_NSP_FUNC_1: PASSED")
        return 1

    def test_gbp_nsp_func_2(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_NSP_FUNC_2: TO CREATE/UPDATE/DELETE/VERIFY a PTG "
            "with NTK-SVC-POLICY with MULTIPLE PTGs\n"
            "TEST_STEPS::\n"
            "Create two NSPolicy Object with non-default params\n"
            "Create PTG using one of the two NSPs\n"
            "Verify the PTG and NSP are reflecting in each other in the DB\n"
            "Update the PTG to use the second NSP\n"
            "Verify the PTG and NSP are reflecting in each other in the DB\n"
            "Update/Revert the PTG so that it refers to the initial NSP\n"
            "Delete all PTG, NSP\n"
            "Verify that PTG and NSPs got deleted\n"
            "##############################################################\n")

        # Testcase work-flow starts
        # Create NSPolicy with non-default attrs
        self._log.info('\n## Step 1: Create two NSPolicy ##\n')
        nsp1_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'nsp',
            'demo_nsp_1',
            network_service_params="type=ip_single,name=vip_ip1,"
                                   "value=self_subnet")
        if nsp1_uuid == 0:
            self._log.info(
                "\n## Step 1A: Create NSPolicy with type=ip_single & "
                "name:self_subnet == Failed")
            return 0
        nsp2_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'nsp',
            'demo_nsp_2',
            network_service_params="type=ip_single,name=vip_ip2,"
                                   "value=self_subnet")
        if nsp2_uuid == 0:
            self._log.info(
                "\n## Step 1B: Create NSPolicy with type=ip_single & "
                "name:self_subnet == Failed")
            return 0
        # Create PTG, referencing one of the two NSPs
        self._log.info(
            "\n## Step 3: Create and Verify PTG referencing one of "
            "the two NSPs")
        uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'group', 'demo_ptg_1', network_service_policy=nsp1_uuid)
        if uuid == 0:
            self._log.info(
                "\n## Step 3A: Create PTG using NSP demo_nsp_1,Failed")
            return 0
        else:
            ptg1_uuid = uuid[0]
        # Verify
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1, 'nsp', nsp1_uuid, policy_target_groups=ptg1_uuid) == 0:
            self._log.info(
                "\n## Step 3B: Verify PTG demo_ptg_1 seen in NSP "
                "demo_nsp_1, Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1, 'group', ptg1_uuid,
                network_service_policy_id=nsp1_uuid) == 0:
            self._log.info(
                "\n## Step 3C: Verify PTG demo_ptg_1 references "
                "NSP demo_nsp_1, Failed")
            return 0
        self._log.info(
            "\n## Step 4: Update and Verify the PTG with the second NSP")
        # Update the PTG with second NSP and Verify
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'group', ptg1_uuid, network_service_policy=nsp2_uuid) == 0:
            self._log.info(
                "\n## Step 4A: Updating NSP attribute of PTG, Failed")
            return 0
        # Verify
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1, 'nsp', nsp1_uuid, policy_target_groups=ptg1_uuid) != 0:
            self._log.info(
                "\n## Step 4B: Verify PTG demo_ptg_1 is NOT seen "
                "in NSP demo_nsp_1, Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1, 'nsp', nsp2_uuid, policy_target_groups=ptg1_uuid) == 0:
            self._log.info(
                "\n## Step 4C: Verify PTG demo_ptg_1 is seen in NSP "
                "demo_nsp_2, Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1, 'group', ptg1_uuid,
                network_service_policy_id=nsp2_uuid) == 0:
            self._log.info(
                "\n## Step 4D: Verify PTG demo_ptg_1 references NSP "
                "demo_nsp_2, Failed")
            return 0
        self._log.info(
            "\n## Step 5: Update/Revert the NSP attr of PTG and Verify")
        # Update the PTG by reverting the NSP to its initial one
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'group', ptg1_uuid, network_service_policy=nsp1_uuid) == 0:
            self._log.info(
                "\n## Step 5A: Reverting the NSP attribute of PTG by "
                "update action, Failed")
            return 0
        # Verify
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1, 'nsp', nsp2_uuid, policy_target_groups=ptg1_uuid) != 0:
            self._log.info(
                "\n## Step 5B: Verify PTG demo_ptg_1 is NOT seen in NSP "
                "demo_nsp_2, Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1, 'nsp', nsp1_uuid, policy_target_groups=ptg1_uuid) == 0:
            self._log.info(
                "\n## Step 5C: Verify PTG demo_ptg_1 is seen in NSP "
                "demo_nsp_1, Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1, 'group', ptg1_uuid,
                network_service_policy_id=nsp1_uuid) == 0:
            self._log.info(
                "\n## Step 5D: Verify PTG demo_ptg_1 references NSP "
                "demo_nsp_1, Failed")
            return 0
        self._log.info(
            "\n## Step 6: Delete and Verify two PTGs each referencing "
            "one of the two NSPs")
        # Delete PTG & NSP
        if self.gbpcfg.gbp_policy_cfg_all(0, 'group', ptg1_uuid) == 0:
            self._log.info("\n## Step 6A: Deletion of PTG,Failed")
            return 0
        nsp_list = [nsp1_uuid, nsp2_uuid]
        for i in range(len(nsp_list)):
            if self.gbpcfg.gbp_policy_cfg_all(0, 'nsp', nsp_list[i]) == 0:
                self._log.info(
                    "\n## Step 6B: Deletion of NSP %s, Failed" %
                    (nsp_list[i]))
                return 0
        # Verify
        for n in range(len(nsp_list)):
            if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                    1, 'nsp', nsp_list[n]) != 0:
                self._log.info("\n## Step 6C: Verify deletion of NSP, Failed")
                return 0
        self._log.info("\n## TESTCASE_GBP_NSP_FUNC_2: PASSED")
        return 1

    def test_gbp_nsp_func_3(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_NSP_FUNC_3: TO CREATE/DELETE/VERIFY "
            "NTK-SVC-POLICY while REFERENCED IN PTG\n"
            "TEST_STEPS::\n"
            "Create NSPolicy Object with non-default params\n"
            "Create PTG referencing the NSP\n"
            "Verify the PTG and NSP are reflecting in each other in the DB\n"
            "Delete and Verify the deletion of referenced NSP fails\n"
            "Delete PTG & NSP, Verify that PTG and NSPs got deleted\n"
            "##############################################################\n")

        # Testcase work-flow starts
        # Create NSPolicy with non-default attrs
        self._log.info(
            '\n## Step 1: Create NSPolicy with non-default params ##\n')
        nsp1_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'nsp',
            'demo_nsp_1',
            network_service_params="type=ip_single,name=vip_ip1,"
                                   "value=self_subnet")
        if nsp1_uuid == 0:
            self._log.info(
                "\n## Step 1A: Create NSPolicy with type=ip_single & "
                "name:self_subnet == Failed")
            return 0
        # Create PTG, referencing one of the two NSPs
        self._log.info(
            "\n## Step 2: Create and Verify PTG referencing the NSP")
        uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'group', 'demo_ptg_1', network_service_policy=nsp1_uuid)
        if uuid == 0:
            self._log.info(
                "\n## Step 2A: Create PTG using NSP demo_nsp_1,Failed")
            return 0
        else:
            ptg1_uuid = uuid[0]
        # Verify
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1, 'nsp', nsp1_uuid, policy_target_groups=ptg1_uuid) == 0:
            self._log.info(
                "\n## Step 2B: Verify PTG demo_ptg_1 seen in NSP demo_nsp_1, "
                "Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1, 'group', ptg1_uuid,
                network_service_policy_id=nsp1_uuid) == 0:
            self._log.info(
                "\n## Step 2C: Verify PTG demo_ptg_1 references "
                "NSP demo_nsp_1, Failed")
            return 0
        # Delete the referenced NSP
        self._log.info(
            "\n## Step 3: Delete the NSP while it is still referenced "
            "in a PTG")
        if self.gbpcfg.gbp_policy_cfg_all(0, 'nsp', nsp1_uuid) != 0:
            self._log.info(
                "\n## Step 3A: Deletion of Referenced NSP DID NOT fail")
            return 0
        # Delete PTG & NSP
        self._log.info("\n## Step 4: Delete PTG followed by NSP and Verify")
        if self.gbpcfg.gbp_policy_cfg_all(0, 'group', ptg1_uuid) == 0:
            self._log.info("\n## Step 4A: Deletion of PTG,Failed")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(0, 'nsp', nsp1_uuid) == 0:
            self._log.info("\n## Step 4B: Deletion of NSP,Failed")
            return 0
        # Verify
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(1, 'nsp', nsp1_uuid) != 0:
            self._log.info("\n## Step 4C: Verify deletion of NSP, Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_NSP_FUNC_3: PASSED")
        return 1


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import sys
import traceback

from neutron_lib import context as n_context
from neutron_lib import rpc as n_rpc
import oslo_messaging as messaging

from gbpservice._i18n import _
from gbpservice.neutron.db import api as db_api
from gbpservice.nfp.common import constants as nfp_constants
from gbpservice.nfp.common import topics as nsf_topics
from gbpservice.nfp.common import utils as nfp_utils
from gbpservice.nfp.core import context as module_context
from gbpservice.nfp.core.event import Event
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import module as nfp_api
from gbpservice.nfp.core.rpc import RpcAgent
from gbpservice.nfp.lib import nfp_context_manager as nfp_ctx_mgr
from gbpservice.nfp.lib import transport
from gbpservice.nfp.orchestrator.db import nfp_db as nfp_db
from gbpservice.nfp.orchestrator.drivers import orchestration_driver
from gbpservice.nfp.orchestrator.openstack import openstack_driver


LOG = nfp_logging.getLogger(__name__)

STOP_POLLING = {'poll': False}
CONTINUE_POLLING = {'poll': True}


def rpc_init(controller, config):
    rpcmgr = RpcHandler(config, controller)
    agent = RpcAgent(
        controller,
        host=config.host,
        topic=nsf_topics.NFP_CONFIGURATOR_NDO_TOPIC,
        manager=rpcmgr)
    controller.register_rpc_agents([agent])


def events_init(controller, config, device_orchestrator):
    events = ['CREATE_NETWORK_FUNCTION_DEVICE', 'DEVICE_SPAWNING',
              'DEVICE_HEALTHY', 'HEALTH_MONITOR_COMPLETE',
              'DEVICE_UP',
              'CONFIGURE_DEVICE', 'CREATE_DEVICE_CONFIGURATION',
              'CONFIGURATION_COMPLETE',
              'DEVICE_CONFIGURED', "DELETE_CONFIGURATION",
              'DELETE_NETWORK_FUNCTION_DEVICE',
              'DELETE_CONFIGURATION_COMPLETED',
              'DEVICE_BEING_DELETED',
              'DEVICE_NOT_REACHABLE',
              'DEVICE_CONFIGURATION_FAILED',
              'PLUG_INTERFACES', 'UNPLUG_INTERFACES',
              'UPDATE_DEVICE_CONFIG_PARAMETERS',
              'DEVICE_CONFIG_PARAMETERS_UPDATED',
              'PERIODIC_HM_DEVICE_REACHABLE',
              'PERIODIC_HM_DEVICE_NOT_REACHABLE',
              'PERFORM_INITIAL_HEALTH_CHECK',
              'PERFORM_PERIODIC_HEALTH_CHECK',
              'PERFORM_CLEAR_HM']
    events_to_register = []
    for event in events:
        events_to_register.append(
            Event(id=event, handler=device_orchestrator))
    controller.register_events(events_to_register)


def nfp_module_init(controller, config):
    events_init(controller, config, DeviceOrchestrator(controller, config))
    rpc_init(controller, config)
    LOG.debug("Device Orchestrator: module_init")


class RpcHandler(object):
    RPC_API_VERSION = '1.0'

    def __init__(self, conf, controller):
        super(RpcHandler, self).__init__()
        self.conf = conf
        self._controller = controller
        self.rpc_event_mapping = {
            nfp_constants.HEALTHMONITOR_RESOURCE: [
                'HEALTH_MONITOR_COMPLETE',
                'DEVICE_NOT_REACHABLE',
                'DEVICE_NOT_REACHABLE',
                'PERIODIC_HM_DEVICE_REACHABLE',
                'PERIODIC_HM_DEVICE_NOT_REACHABLE', ],
            nfp_constants.GENERIC_CONFIG: [
                'DEVICE_CONFIGURED',
                'DELETE_CONFIGURATION_COMPLETED',
                'DEVICE_CONFIGURATION_FAILED'],
        }

    def _log_event_created(self, event_id, event_data):
        NFD = event_data.get('network_function_device_id')
        NF = event_data.get('network_function_id')
        NFI = event_data.get('network_function_instance_id')

        if NFD and NF and NFI:
            LOG.info("Created event %(event_name)s with"
                     " NF:%(nf)s ,NFI:%(nfi)s and NFD:%(nfd)s",
                     {'event_name': event_id,
                      'nf': NF,
                      'nfi': NFI,
                      'nfd': NFD})
        else:
            LOG.info("Created event %(event_name)s ",
                     {'event_name': event_id})

    def _create_event(self, event_id, event_data=None, key=None,
                      is_poll_event=False, original_event=False, max_times=10):
        if is_poll_event:
            ev = self._controller.new_event(
                id=event_id, data=event_data,
                serialize=original_event.sequence,
                binding_key=original_event.binding_key,
                key=original_event.desc.uuid)
            LOG.debug("poll event started for %s", ev.id)
            self._controller.poll_event(ev, max_times=10)
        else:
            ev = self._controller.new_event(
                id=event_id,
                key=key,
                data=event_data)
            self._controller.post_event(ev)
        self._log_event_created(event_id, event_data)

    def handle_periodic_hm_resource(self, result):
        if result == nfp_constants.SUCCESS:
            event_id = self.rpc_event_mapping[
                nfp_constants.HEALTHMONITOR_RESOURCE][3]
        else:
            event_id = self.rpc_event_mapping[
                nfp_constants.HEALTHMONITOR_RESOURCE][4]
        return event_id

    # RPC APIs status notification from Configurator
    def network_function_notification(self, context, notification_data):
        nfp_context = module_context.init()
        info = notification_data.get('info')
        responses = notification_data.get('notification')
        request_info = info.get('context')
        operation = request_info.get('operation')
        logging_context = request_info.get('logging_context', {})
        # nfp_context = request_info.get('nfp_context')
        nfp_context['log_context'] = logging_context
        if 'nfp_context' in request_info:
            nfp_context.update(request_info['nfp_context'])

        for response in responses:
            resource = response.get('resource')
            data = response.get('data')
            result = data.get('status_code')
            if resource not in [nfp_constants.HEALTHMONITOR_RESOURCE,
                                nfp_constants.PERIODIC_HM]:
                resource = nfp_constants.GENERIC_CONFIG

            is_delete_request = True if operation == 'delete' else False

            if resource == nfp_constants.PERIODIC_HM:
                event_id = self.handle_periodic_hm_resource(result)
                break

            if is_delete_request:
                event_id = self.rpc_event_mapping[resource][1]
            else:
                event_id = self.rpc_event_mapping[resource][0]

            if result.lower() != 'success':
                LOG.info("RPC Handler response data:%(data)s",
                         {'data': data})
                if is_delete_request:
                    # Ignore any deletion errors, generate SUCCESS event
                    event_id = self.rpc_event_mapping[resource][1]
                else:
                    event_id = self.rpc_event_mapping[resource][2]
                break

        nf_id = request_info.pop('nf_id')
        nfi_id = request_info.pop('nfi_id')
        nfd_id = request_info.pop('nfd_id')
        request_info['network_function_id'] = nf_id
        request_info['network_function_instance_id'] = nfi_id
        request_info['network_function_device_id'] = nfd_id
        event_data = request_info
        event_data['id'] = request_info['network_function_device_id']

        key = nf_id
        self._create_event(event_id=event_id,
                           event_data=event_data,
                           key=key)


class DeviceOrchestrator(nfp_api.NfpEventHandler):
    """device Orchestrator For Network Services

    This class handles the orchestration of Network Function Device lifecycle.
    It deals with physical service resources - Network Devices. This module
    interacts with Service Orchestrator and Configurator. Service Orchestrator
    sends device create/delete request, Device orchestrator sends/receieves
    RPC to/from configurator to create generic config. Device Orchestrator
    loads drivers specified in config file, and selects corresponding drivers
    based on service vendor.

    Workflow for create:
    1) Service Orchestarator calls Device Orcehstrator(NDO) for creating new
    device, create_network_function_device gets called in the context of
    event handler. This method checks with drivers for device sharing, if
    device sharing supported then request the driver to plug the
    interfaces(step-3) else request driver to create a new device and poll
    for its status(Here a DB entry is created with status as PENDING_CREATE).
    2) If the driver returns status as ACTIVE or ERROR, the poll event is
    stopped, if it returns any other status, the poll event is continued.
    2) In case of new device creation, once device become ACTIVE, NDO plug
    the interfaces.
    3) After plugging the interfaces NDO sends RPC call to configurator for
    creating generic config.
    4) Rpc Handler receives notification API from configurator, In case of
    success update DB with status as ACTIVE and create event DEVICE_CREATED
    for Service Orchestrator. In case of any error, create event
    DEVICE_CREATE_FAILED and update DB as ERROR.

    """

    def __init__(self, controller, config):
        self._controller = controller
        self.config = config
        self.nsf_db = nfp_db.NFPDbBase()
        self.gbpclient = openstack_driver.GBPClient(config)
        self.keystoneclient = openstack_driver.KeystoneClient(config)

        neutron_context = n_context.get_admin_context()
        self.configurator_rpc = NDOConfiguratorRpcApi(neutron_context,
                                                      self.config)

        self.status_map = {
            'INIT': 'Created Network Service Device with status INIT.',
            'PENDING_CREATE': '',
            'DEVICE_SPAWNING': ('Creating NFD, launched the new device, ' +
                                'polling on its status'),
            'DEVICE_UP': 'Device is UP/ACTIVE',
            'PERFORM_HEALTH_CHECK': 'perform health check of service vm',
            'HEALTH_CHECK_PENDING': ('Device health check is going on ' +
                                     ' through configurator'),
            'HEALTH_CHECK_COMPLETED': 'Health check succesfull for device',
            'INTERFACES_PLUGGED': 'Interfaces Plugging successfull',
            'PENDING_CONFIGURATION_CREATE': ('Started configuring device ' +
                                             'for routes, license, etc'),
            'DEVICE_READY': 'Device is ready to use',
            'ACTIVE': 'Device is Active.',
            'DEVICE_NOT_UP': 'Device not became UP/ACTIVE',
        }
        self.orchestration_driver = orchestration_driver.OrchestrationDriver(
            self.config)

    @property
    def db_session(self):
        return db_api.get_writer_session()

    def event_method_mapping(self, event_id):
        event_handler_mapping = {
            "CREATE_NETWORK_FUNCTION_DEVICE": (
                self.create_network_function_device),
            "PERFORM_INITIAL_HEALTH_CHECK": self.perform_initial_health_check,
            "PERFORM_PERIODIC_HEALTH_CHECK":
                self.perform_periodic_health_check,
            "PERFORM_CLEAR_HM": self.perform_clear_hm,
            "DEVICE_UP": self.device_up,
            "PLUG_INTERFACES": self.plug_interfaces_fast,
            "DEVICE_HEALTHY": self.plug_interfaces,
            "HEALTH_MONITOR_COMPLETE": self.health_monitor_complete,
            "CONFIGURE_DEVICE": self.configure_device,
            "CREATE_DEVICE_CONFIGURATION": self.create_device_configuration,
            "CONFIGURATION_COMPLETE": self.configuration_complete,
            "DEVICE_CONFIGURED": self.device_configuration_complete,

            "DELETE_NETWORK_FUNCTION_DEVICE": (
                self.delete_network_function_device),
            "DELETE_CONFIGURATION_COMPLETED": (
                self.delete_configuration_complete),
            "UNPLUG_INTERFACES": self.unplug_interfaces,
            "DELETE_DEVICE": self.delete_device,
            "DELETE_CONFIGURATION": self.delete_device_configuration,
            "DEVICE_NOT_REACHABLE": self.handle_device_not_reachable,
            "PERIODIC_HM_DEVICE_REACHABLE": (
                self.periodic_hm_handle_device_reachable),
            "PERIODIC_HM_DEVICE_NOT_REACHABLE": (
                self.periodic_hm_handle_device_not_reachable),
            "PLUG_INTERFACE_FAILED": self.handle_plug_interface_failed,
            "DEVICE_CONFIGURATION_FAILED": self.handle_device_config_failed,
            "DEVICE_ERROR": self.handle_device_create_error,
            "DEVICE_NOT_UP": self.handle_device_not_up,
            "DRIVER_ERROR": self.handle_driver_error,
            'UPDATE_DEVICE_CONFIG_PARAMETERS': self.update_config_params,
            'DEVICE_CONFIG_PARAMETERS_UPDATED': (
                self.device_configuration_updated)
        }
        if event_id not in event_handler_mapping:
            raise Exception(_("Invalid event ID"))
        else:
            return event_handler_mapping[event_id]

    def handle_event(self, event):
        try:
            event_data = event.context
            NFD = event_data.get('network_function_device_id')
            NF = event_data.get('network_function_id')
            NFI = event_data.get('network_function_instance_id')

            if NFD and NF and NFI:
                LOG.info("Received event %(event_name)s with "
                         "NF:%(nf)s ,NFI:%(nfi)s and NFD:%(nfd)s",
                         {'event_name': event.id,
                          'nf': NF,
                          'nfi': NFI,
                          'nfd': NFD})
            else:
                LOG.info("Received event %(event_name)s ",
                         {'event_name': event.id})
            event_handler = self.event_method_mapping(event.id)
            event_handler(event)
        except Exception as e:
            LOG.error("error in processing event: %(event_id)s for "
                      "event data %(event_data)s. error: %(error)s",
                      {'event_id': event.id, 'event_data': event.data,
                       'error': e})
            _, _, tb = sys.exc_info()
            traceback.print_tb(tb)
            raise e

    def handle_exception(self, event, exception):
        return ExceptionHandler.handle(self, event, exception)

    # Helper functions
    def _log_event_created(self, event_id, event_data):
        network_function_instance = event_data.get('network_function_instance')
        if network_function_instance:
            nf = network_function_instance.get('network_function_id')
            nfi = network_function_instance.get('id')
        else:
            nf = None
            nfi = None
        if nf and nfi:
            LOG.info("Created event %(event_name)s with NF:%(nf)s and "
                     "NFI:%(nfi)s ",
                     {'event_name': event_id,
                      'nf': nf,
                      'nfi': nfi})
        else:
            LOG.info("Created event %(event_name)s ",
                     {'event_name': event_id})

    def _create_event(self, event_id, event_data=None,
                      is_poll_event=False, original_event=False,
                      is_internal_event=False, max_times=10):
        if not is_internal_event:
            if is_poll_event:
                ev = self._controller.new_event(
                    id=event_id, data=event_data,
                    serialize=original_event.sequence,
                    binding_key=original_event.binding_key,
                    key=original_event.desc.uuid)
                LOG.debug("poll event started for %s", ev.id)
                self._controller.poll_event(ev, max_times=max_times)
            else:
                ev = self._controller.new_event(
                    id=event_id,
                    data=event_data)
                self._controller.post_event(ev)
            nfp_context = module_context.get()
            self._log_event_created(event_id, nfp_context)
        else:
            # Same module API, so calling corresponding function directly.
            nfp_context = module_context.get()
            event = self._controller.new_event(
                id=event_id,
                data=event_data,
                context=nfp_context)
            self.handle_event(event)

    def _release_cnfd_lock(self, device):
        nf_id = device['network_function_id']
        nfi_id = device['network_function_instance_id']
        ev = self._controller.new_event(
            id='CREATE_NETWORK_FUNCTION_DEVICE',
            data=device, key=nf_id + nfi_id)
        if device.get('binding_key'):
            ev.binding_key = device.get('binding_key')
            LOG.debug("Releasing tenant based lock for "
                      "CREATE_NETWORK_FUNCTION_DEVICE event with binding "
                      "key: %s", ev.binding_key)
        self._controller.event_complete(ev)

    def event_cancelled(self, ev, reason):
        LOG.info("Poll event %(event_id)s cancelled.",
                 {'event_id': ev.id})

        if ev.id == 'DEVICE_SPAWNING':
            LOG.info("Device is not up still after 10secs of launch")
            # create event DEVICE_NOT_UP
            device = self._prepare_failure_case_device_data(ev.data)
            self._create_event(event_id='DEVICE_NOT_UP',
                               event_data=device,
                               is_internal_event=True)
            self._update_network_function_device_db(device,
                                                    'DEVICE_NOT_UP')
        if ev.id == 'DEVICE_BEING_DELETED':
            LOG.info("Device is not deleted completely."
                     " Continuing further cleanup of resources."
                     " Possibly there could be stale port resources"
                     " on Compute")
            device = ev.data
            orchestration_driver = self._get_orchestration_driver(
                device['service_details']['service_vendor'])
            device_id = device['id']
            del device['id']
            nf_id = device['network_function_id']
            orchestration_driver.delete_network_function_device(device)
            self._delete_network_function_device_db(device_id, device)
            dnfd_event = (
                self._controller.new_event(id='DELETE_NETWORK_FUNCTION_DEVICE',
                                           key=nf_id,
                                           binding_key=nf_id,
                                           desc_dict=device.get(
                                               'event_desc')))
            self._controller.event_complete(dnfd_event, result='FAILED')

    def _update_device_status(self, device, state, status_desc=None):
        device['status'] = state
        if status_desc:
            device['status_description'] = status_desc
        else:
            device['status_description'] = self.status_map.get(state)

    def _get_port(self, port_id):
        with nfp_ctx_mgr.DbContextManager:
            return self.nsf_db.get_port_info(self.db_session, port_id)

    def _get_ports(self, port_ids):
        data_ports = []
        for port_id in port_ids:
            with nfp_ctx_mgr.DbContextManager:
                port_info = self.nsf_db.get_port_info(self.db_session,
                                                      port_id)
            data_ports.append(port_info)
            return data_ports

    def _create_network_function_device_db(self, device_info, state):

        self._update_device_status(device_info, state)
        # (ashu) driver should return device_id as vm_id
        device_id = device_info.pop('id')
        device_info['id'] = device_id
        device_info['reference_count'] = 0
        device_info['interfaces_in_use'] = 0
        with nfp_ctx_mgr.DbContextManager:
            device = self.nsf_db.create_network_function_device(
                self.db_session,
                device_info)
        return device

    def _update_network_function_device_db(self, device, state,
                                           status_desc=''):
        self._update_device_status(device, state, status_desc)
        updated_device = copy.deepcopy(device)
        updated_device.pop('reference_count', None)
        updated_device.pop('interfaces_in_use', None)
        with nfp_ctx_mgr.DbContextManager:
            self.nsf_db.update_network_function_device(self.db_session,
                                                       updated_device['id'],
                                                       updated_device)
        device.update(updated_device)

    def _delete_network_function_device_db(self, device_id, device):
        with nfp_ctx_mgr.DbContextManager:
            self.nsf_db.delete_network_function_device(self.db_session,
                                                       device_id)

    def _get_network_function_info(self, device_id):
        nfi_filters = {'network_function_device_id': [device_id]}
        with nfp_ctx_mgr.DbContextManager:
            network_function_instances = (
                self.nsf_db.get_network_function_instances(self.db_session,
                                                           nfi_filters))
            network_function_ids = [nf['network_function_id']
                                    for nf in network_function_instances]
            network_functions = (
                self.nsf_db.get_network_functions(
                    self.db_session,
                    {'id': network_function_ids}))
            return network_functions

    def _get_network_function_devices(self, filters=None):
        with nfp_ctx_mgr.DbContextManager:
            network_function_devices = (
                self.nsf_db.get_network_function_devices(self.db_session,
                                                         filters))
        for device in network_function_devices:
            mgmt_port_id = device.pop('mgmt_port_id')
            mgmt_port_id = self._get_port(mgmt_port_id)
            device['mgmt_port_id'] = mgmt_port_id

            network_functions = (
                self._get_network_function_info(device['id']))
            device['network_functions'] = network_functions
        return network_function_devices

    def _increment_device_ref_count(self, device):
        with nfp_ctx_mgr.DbContextManager:
            self.nsf_db.increment_network_function_device_count(
                self.db_session,
                device['id'],
                'reference_count')
        device['reference_count'] += 1

    def _decrement_device_ref_count(self, device):
        with nfp_ctx_mgr.DbContextManager:
            self.nsf_db.decrement_network_function_device_count(
                self.db_session,
                device['id'],
                'reference_count')
        device['reference_count'] -= 1

    def _increment_device_interface_count(self, device):
        with nfp_ctx_mgr.DbContextManager:
            self.nsf_db.increment_network_function_device_count(
                self.db_session,
                device['id'],
                'interfaces_in_use',
                len(device['ports']))

        device['interfaces_in_use'] += len(device['ports'])

    def _decrement_device_interface_count(self, device):
        with nfp_ctx_mgr.DbContextManager:
            self.nsf_db.decrement_network_function_device_count(
                self.db_session,
                device['id'],
                'interfaces_in_use',
                len(device['ports']))

        device['interfaces_in_use'] -= len(device['ports'])

    def _get_orchestration_driver(self, service_vendor):
        return self.orchestration_driver

    def _get_device_data(self, nfd_request):

        device_data = {}
        network_function = nfd_request.get('network_function')
        network_function_instance = nfd_request['network_function_instance']
        service_details = nfd_request['service_details']
        device_data['name'] = network_function_instance['name']
        device_data['share_existing_device'] = (
            nfd_request.get('share_existing_device'))
        device_data['management_network_info'] = (
            nfd_request.get('management_network_info'))

        if network_function:
            device_data['network_function_id'] = network_function['id']
            device_data['service_chain_id'] = (
                network_function['service_chain_id'])

        device_data['network_function_instance_id'] = (
            network_function_instance['id'])
        device_data['tenant_id'] = network_function_instance['tenant_id']

        nsi_port_info = []
        for port_id in network_function_instance.pop('port_info'):
            with nfp_ctx_mgr.DbContextManager:
                port_info = self.nsf_db.get_port_info(self.db_session,
                                                      port_id)
            nsi_port_info.append(port_info)

        device_data['ports'] = nsi_port_info

        device_data['service_details'] = service_details
        if nsi_port_info[0]['port_model'] == nfp_constants.GBP_PORT:
            device_data['service_details']['network_mode'] = (
                nfp_constants.GBP_MODE)
        else:
            device_data['service_details']['network_mode'] = (
                nfp_constants.NEUTRON_MODE)
        device_data['service_vendor'] = service_details['service_vendor']

        return device_data

    def _get_nsf_db_resource(self, resource_name, resource_id):
        db_method = getattr(self.nsf_db, 'get_' + resource_name)
        return db_method(self.db_session, resource_id)

    def _update_device_data(self, device, device_data):
        device.update(device_data)
        return device

    def _make_ports_dict(self, consumer, provider, port_type):

        t_ports = []
        for ptg in [consumer, provider]:
            if (port_type in ptg.keys()) and ptg[port_type]:
                t_ports.append({
                    'id': ptg[port_type].get('id'),
                    'port_classification': ptg.get(
                        'port_classification'),
                    'port_model': ptg.get('port_model')
                })
        return t_ports

    def _prepare_device_data_from_nfp_context(self, nfp_context):
        device_data = {}

        network_function = nfp_context['network_function']
        network_function_instance = nfp_context['network_function_instance']
        service_details = nfp_context['service_details']

        device_data['token'] = nfp_context[
            'resource_owner_context']['admin_token']
        device_data['admin_tenant_id'] = nfp_context[
            'resource_owner_context']['admin_tenant_id']
        device_data['name'] = network_function_instance['name']
        device_data['share_existing_device'] = nfp_context[
            'share_existing_device']

        management_network_info = {
            'id': nfp_context['management_ptg_id'],
            'port_model': nfp_constants.GBP_NETWORK
        }

        consumer = nfp_context['consumer']
        provider = nfp_context['provider']

        ports = self._make_ports_dict(nfp_context.get(
            'explicit_consumer', consumer), provider, 'pt')

        device_data['provider_name'] = provider['ptg']['name']
        device_data['management_network_info'] = management_network_info

        device_data['network_function_id'] = network_function['id']
        device_data['service_chain_id'] = network_function['service_chain_id']
        device_data[
            'network_function_instance_id'] = network_function_instance['id']
        device_data['tenant_id'] = network_function_instance['tenant_id']
        device_data['ports'] = ports
        device_data['service_details'] = service_details
        device_data['service_details']['network_mode'] = nfp_constants.GBP_MODE
        device_data['service_vendor'] = service_details['service_vendor']
        device_data['server_grp_id'] = nfp_context.get('server_grp_id')
        device_data['interfaces_to_attach'] = (
            nfp_context.get('interfaces_to_attach'))

        if nfp_context.get('files'):
            device_data['files'] = nfp_context['files']
        if nfp_context.get('user_data'):
            device_data['user_data'] = nfp_context['user_data']
        return device_data

    def _create_nfd_entry(self, nfp_context, driver_device_info,
                          device_data, service_details):
        nfp_context['provider_metadata'] = driver_device_info.get(
            'provider_metadata')
        # Update nfp_context management with newly created mgmt port
        management = nfp_context['management']
        management['port'] = driver_device_info[
            'mgmt_neutron_port_info']['neutron_port']
        management['port']['ip_address'] = management[
            'port']['fixed_ips'][0]['ip_address']
        management['subnet'] = driver_device_info[
            'mgmt_neutron_port_info']['neutron_subnet']

        # Update newly created device with required params
        device = self._update_device_data(driver_device_info, device_data)
        device['network_function_device_id'] = device['id']

        # check for any explicit interface and its type.
        for interface in nfp_context.get('explicit_interfaces', []):
            if interface['type'] == 'gateway':
                device['gateway_port'] = interface['port']

        name = '%s_%s_%s_%s' % (
            device['provider_name'],
            service_details['service_type'],
            nfp_context['resource_owner_context']['tenant_name'][:6],
            device['network_function_device_id'][:3])
        device['name'] = name
        # Create DB entry with status as DEVICE_SPAWNING
        network_function_device = (
            self._create_network_function_device_db(device,
                                                    'DEVICE_SPAWNING'))

        nfp_context['network_function_device'] = network_function_device
        return device

    def _update_nfp_context_with_ports(self, nfp_context, device):
        # REVISIT(mak) Wrong but nfp_db method needs in this format
        network_function_device = nfp_context['network_function_device']
        network_function_device['mgmt_port_id'] = device['mgmt_port_id']

    def _post_create_nfd_events(self, event, nfp_context, device):

        nfp_context['event_desc'] = event.desc.to_dict()
        # Updating nfi with nfd_id before device spawning
        # to stop orchestration to move further.
        nfi = {
            'network_function_device_id': device['id'],
        }
        with nfp_ctx_mgr.DbContextManager:
            nfi = self.nsf_db.update_network_function_instance(
                self.db_session,
                device['network_function_instance_id'], nfi)
        # This event is act as a dummy event for nfp,
        # for non-hotplug sharing it will be used
        self._create_event(event_id='DEVICE_CREATED',
                           event_data=device)

        self._create_event(event_id='DEVICE_SPAWNING',
                           event_data=nfp_context,
                           is_poll_event=True,
                           original_event=event,
                           max_times=nfp_constants.DEVICE_SPAWNING_MAXRETRY)

    # Create path
    def create_network_function_device(self, event):
        """ Returns device instance for a new service

        This method either returns existing device which could be reused for a
        new service or it creates new device instance
        """

        nfp_context = event.context
        nfd_request = self._prepare_failure_case_device_data(nfp_context)
        service_details = nfp_context['service_details']

        LOG.info("Received event CREATE NETWORK FUNCTION "
                 "DEVICE request.")

        orchestration_driver = self._get_orchestration_driver(
            service_details['service_vendor'])

        device_data = self._prepare_device_data_from_nfp_context(nfp_context)

        LOG.info("Creating new device:%(device)s",
                 {'device': nfd_request})
        device_data['volume_support'] = (
            self.config.device_orchestrator.volume_support)
        device_data['volume_size'] = (
            self.config.device_orchestrator.volume_size)
        device_data['explicit_interfaces'] = nfp_context.get(
            'explicit_interfaces', [])
        driver_device_info = (
            orchestration_driver.create_network_function_device(
                device_data))
        if not driver_device_info:
            LOG.info("Device creation failed")
            self._create_event(event_id='DEVICE_ERROR',
                               event_data=nfd_request,
                               is_internal_event=True)
            self._controller.event_complete(event)
            return None

        device = self._create_nfd_entry(nfp_context, driver_device_info,
                                        device_data, service_details)
        self._increment_device_ref_count(device)
        self._increment_device_interface_count(device)
        nfd_id = device.get('network_function_device_id',
                            '-') if device else '-'
        nfp_context['log_context']['nfd_id'] = nfd_id
        self._update_nfp_context_with_ports(nfp_context, driver_device_info)

        self._post_create_nfd_events(event, nfp_context, device)

    def _post_device_up_event_graph(self, nfp_context):
        nf_id = nfp_context['network_function']['id']
        nfi_id = nfp_context['network_function_instance']['id']
        du_event = self._controller.new_event(id="DEVICE_UP",
                                              key=nf_id + nfi_id)

        hc_event = self._controller.new_event(
            id="PERFORM_INITIAL_HEALTH_CHECK",
            key=nf_id + nfi_id)

        plug_int_event = self._controller.new_event(id="PLUG_INTERFACES",
                                                    key=nf_id + nfi_id)
        GRAPH = ({
            du_event: [hc_event, plug_int_event]})

        self._controller.post_graph(
            GRAPH, du_event, graph_str='HEALTH_MONITOR_GRAPH')

    @nfp_api.poll_event_desc(event='DEVICE_SPAWNING',
                             spacing=nfp_constants.DEVICE_SPAWNING_SPACING)
    def check_device_is_up(self, event):
        nfp_context = event.context

        service_details = nfp_context['service_details']
        network_function_device = nfp_context['network_function_device']
        token = nfp_context['resource_owner_context']['admin_token']
        tenant_id = nfp_context['resource_owner_context']['admin_tenant_id']

        device = {
            'token': token,
            'tenant_id': tenant_id,
            'id': network_function_device['id'],
            'service_details': service_details}

        orchestration_driver = self._get_orchestration_driver(
            service_details['service_vendor'])

        is_device_up = (
            orchestration_driver.get_network_function_device_status(device))

        if is_device_up == nfp_constants.ACTIVE:
            LOG.info("Device with NFD:%(id)s came up for "
                     "tenant:%(tenant)s ",
                     {'id': network_function_device['id'],
                      'tenant': tenant_id})
            self._post_device_up_event_graph(nfp_context)

            return STOP_POLLING
        elif is_device_up == nfp_constants.ERROR:
            # create event DEVICE_NOT_UP
            device = self._prepare_failure_case_device_data(nfp_context)
            self._create_event(event_id='DEVICE_NOT_UP',
                               event_data=device,
                               is_internal_event=True)
            self._update_network_function_device_db(device,
                                                    'DEVICE_NOT_UP')
            return STOP_POLLING
        else:
            # Continue polling until device status became ACTIVE/ERROR.
            return CONTINUE_POLLING

    def _post_configure_device_graph(self, nfp_context, serialize=False):
        nf_id = nfp_context['network_function']['id']
        nfi_id = nfp_context['network_function_instance']['id']
        sc_instance_id = nfp_context['service_chain_instance']['id']
        binding_key = nfp_context['service_details'][
            'service_vendor'].lower() + nf_id
        device_configure_event = self._controller.new_event(
            id='CREATE_DEVICE_CONFIGURATION',
            key=nf_id,
            serialize=serialize,
            binding_key=binding_key)
        check_heat_config = self._controller.new_event(
            id='SEND_USER_CONFIG',
            key=nf_id)
        user_config_event = self._controller.new_event(
            id='INITIATE_USER_CONFIG',
            key=nf_id,
            serialize=serialize,
            binding_key=binding_key)
        device_configured_event = self._controller.new_event(
            id='CONFIGURATION_COMPLETE',
            key=nf_id,
            serialize=serialize,
            binding_key=sc_instance_id)
        device_periodic_hm_event = self._controller.new_event(
            id='PERFORM_PERIODIC_HEALTH_CHECK',
            key=nf_id + nfi_id)

        # Start periodic health monitor after device configuration

        GRAPH = ({
            device_periodic_hm_event: [device_configured_event],
            device_configured_event: [device_configure_event,
                                      user_config_event],
            user_config_event: [check_heat_config]})

        self._controller.post_graph(GRAPH, device_periodic_hm_event,
                                    graph_str='DEVICE_CONFIGURATION_GRAPH')

    def device_up(self, event, serialize_config=False):
        nfp_context = event.context

        # Get the results of PLUG_INTERFACES & PERFORM_INITIAL_HEALTH_CHECK
        # events results.
        nf_id = nfp_context['network_function']['id']
        nfi_id = nfp_context['network_function_instance']['id']
        device = self._prepare_failure_case_device_data(nfp_context)
        # Get the results of PLUG_INTERFACES & PERFORM_INITIAL_HEALTH_CHECK
        # events results.
        results = event.result
        for result in results:
            if result.result.lower() != 'success':
                # Release CNFD Event lock
                self._release_cnfd_lock(device)
                self._create_event(event_id='DEVICE_CREATE_FAILED',
                                   event_data=device)
                return self._controller.event_complete(event, result='FAILED')

        network_function_device = nfp_context['network_function_device']

        nfd_id = '-'
        if network_function_device:
            nfd_id = network_function_device.get('id', '-')
        nfp_context['log_context']['nfd_id'] = nfd_id
        # Update NFI to ACTIVE State
        nfi = {
            'status': nfp_constants.ACTIVE}
        nfi = self.nsf_db.update_network_function_instance(
            self.db_session, nfi_id, nfi)
        self._update_network_function_device_db(
            network_function_device, nfp_constants.ACTIVE)

        LOG.info(
            "Configuration completed for device with NFD:%(device_id)s. "
            "Updated DB status to ACTIVE.",
            {'device_id': network_function_device['id']})
        LOG.debug("Device detail:%s",
                  network_function_device)
        # Release CNFD Event lock
        self._release_cnfd_lock(device)
        self._post_configure_device_graph(nfp_context,
                                          serialize=serialize_config)
        event.key = nf_id + nfi_id
        self._controller.event_complete(event)

    def prepare_health_check_device_info(self, event, periodicity):

        nfp_context = event.context

        service_details = nfp_context['service_details']
        network_function_device = nfp_context['network_function_device']
        network_function = nfp_context['network_function']
        network_function_instance = nfp_context['network_function_instance']
        mgmt_ip_address = nfp_context['management']['port']['ip_address']
        tenant_id = nfp_context['resource_owner_context']['admin_tenant_id']

        # The driver tells which protocol / port to monitor ??
        orchestration_driver = self._get_orchestration_driver(
            service_details['service_vendor'])
        nfp_context['event_desc'] = event.desc.to_dict()
        device = {
            'id': network_function_device['id'],
            'tenant_id': tenant_id,
            'mgmt_ip_address': mgmt_ip_address,
            'service_details': service_details,
            'network_function_id': network_function['id'],
            'periodicity': periodicity,
            'network_function_instance_id': network_function_instance['id'],
            'nfp_context': {'event_desc': nfp_context['event_desc'],
                            'id': event.id, 'key': event.key},
        }
        return device, orchestration_driver

    def perform_clear_hm(self, event):
        nfp_context = event.data
        network_function = nfp_context['network_function']
        service_details = nfp_context['service_details']
        orchestration_driver = self._get_orchestration_driver(
            service_details['service_vendor'])
        nfp_context['event_desc'] = event.desc.to_dict()
        device = {
            'id': nfp_context['network_function_device_id'],
            'tenant_id': nfp_context['tenant_id'],
            'mgmt_ip_address': nfp_context['mgmt_ip_address'],
            'service_details': service_details,
            'network_function_id': network_function['id'],
            'network_function_instance_id': nfp_context[
                'network_function_instance_id'],
            'nfp_context': {'event_desc': nfp_context['event_desc'],
                            'id': event.id, 'key': event.key},
        }
        clear_hm_req = (
            orchestration_driver.get_network_function_device_config(
                device, nfp_constants.HEALTHMONITOR_RESOURCE))
        if not clear_hm_req:
            self._controller.event_complete(event, result="FAILED")
            return None

        self.configurator_rpc.delete_network_function_device_config(
            device,
            clear_hm_req)
        LOG.debug("Clear HM RPC sent to configurator for device: "
                  "%s with parameters: %s", (
                      device['id'], clear_hm_req))
        self._controller.event_complete(event, result="SUCCESS")

    def perform_periodic_health_check(self, event):
        event_results = event.result
        for result in event_results:
            if result.result.lower() != "success":
                return self._controller.event_complete(event, result="FAILED")

        device, orchestration_driver = (
            self.prepare_health_check_device_info(event,
                                                  nfp_constants.FOREVER))
        hm_req = (
            orchestration_driver.get_network_function_device_config(
                device, nfp_constants.HEALTHMONITOR_RESOURCE))
        if not hm_req:
            self._controller.event_complete(event, result="FAILED")
            return None

        self.configurator_rpc.create_network_function_device_config(device,
                                                                    hm_req)
        LOG.debug("Health Check RPC sent to configurator for device: "
                  "%s with health check parameters: %s", (
                      device['id'], hm_req))
        self._controller.event_complete(event, result="SUCCESS")

    def perform_initial_health_check(self, event):
        device, orchestration_driver = (
            self.prepare_health_check_device_info(event,
                                                  nfp_constants.INITIAL))
        hm_req = (
            orchestration_driver.get_network_function_device_config(
                device, nfp_constants.HEALTHMONITOR_RESOURCE))
        if not hm_req:
            self._controller.event_complete(event, result="FAILED")
            return None
        self.configurator_rpc.create_network_function_device_config(device,
                                                                    hm_req)
        LOG.debug("Health Check RPC sent to configurator for device: "
                  "%s with health check parameters: %s", (
                      device['id'], hm_req))

    def _get_service_type(self, service_profile_id):
        with nfp_ctx_mgr.KeystoneContextManager as kcm:
            admin_token = kcm.retry(
                self.keystoneclient.get_admin_token, tries=3)
        with nfp_ctx_mgr.GBPContextManager as gcm:
            service_profile = gcm.retry(self.gbpclient.get_service_profile,
                                        admin_token, service_profile_id)
        return service_profile['service_type'].lower()

    def _prepare_device_data(self, device_info):
        network_function_id = device_info['network_function_id']
        network_function_device_id = device_info['network_function_device_id']
        network_function_instance_id = (
            device_info['network_function_instance_id'])

        network_function = self._get_nsf_db_resource(
            'network_function',
            network_function_id)
        network_function_device = self._get_nsf_db_resource(
            'network_function_device',
            network_function_device_id)
        network_function_instance = self._get_nsf_db_resource(
            'network_function_instance',
            network_function_instance_id)

        with nfp_ctx_mgr.KeystoneContextManager as kcm:
            admin_token = kcm.retry(
                self.keystoneclient.get_admin_token, tries=3)
        with nfp_ctx_mgr.GBPContextManager as gcm:
            service_profile = gcm.retry(
                self.gbpclient.get_service_profile,
                admin_token,
                network_function['service_profile_id'])
        service_details = transport.parse_service_flavor_string(
            service_profile['service_flavor'])

        device_info.update({
            'network_function_instance': network_function_instance})
        device_info.update({'id': network_function_device_id})
        service_details.update({'service_type': self._get_service_type(
            network_function['service_profile_id'])})
        device_info.update({'service_details': service_details})

        device = self._get_device_data(device_info)
        device = self._update_device_data(device, network_function_device)

        mgmt_port_id = network_function_device.pop('mgmt_port_id')
        mgmt_port_id = self._get_port(mgmt_port_id)
        device['mgmt_port_id'] = mgmt_port_id
        device['network_function_id'] = network_function_id

        return device

    def _prepare_device_data_fast(self, network_function_details):
        network_function = network_function_details['network_function']
        network_function_device = network_function_details[
            'network_function_device']
        admin_token = network_function_details['admin_token']
        service_profile = network_function_details['service_profile']
        service_details = network_function_details['service_details']
        service_details.update(
            {'service_type': service_profile['service_type']})
        device = self._get_device_data(network_function_details)
        device = self._update_device_data(device, network_function_device)
        mgmt_port_id = network_function_device.pop('mgmt_port_id')
        mgmt_port_id = self._get_port(mgmt_port_id)
        device['mgmt_port_id'] = mgmt_port_id
        device['network_function_id'] = network_function['id']
        device['network_function_device_id'] = (
            network_function_device['id'])
        device['token'] = admin_token
        device['tenant_id'] = (
            network_function_details['admin_tenant_id'])
        device['service_profile'] = service_profile
        return device

    def health_monitor_complete(self, event, result='SUCCESS'):
        nfp_context = event.data['nfp_context']
        # device = nfp_context['network_function_device']
        # network_function = nfp_context['network_function']

        # Invoke event_complete for original event which is
        # PERFORM_INITIAL_HEALTH_CHECK
        event_desc = nfp_context.pop('event_desc', None)
        nfp_context.pop('id', None)
        key = nfp_context.pop('key', None)
        self._controller.event_complete(event)
        new_event = self._controller.new_event(
            id="PERFORM_INITIAL_HEALTH_CHECK",
            key=key, desc_dict=event_desc)
        self._controller.event_complete(new_event, result=result)

    def plug_interfaces(self, event, is_event_call=True):
        if is_event_call:
            device_info = event.data
        else:
            device_info = event
        # Get event data, as configurator sends back only request_info, which
        # contains nf_id, nfi_id, nfd_id.
        device = self._prepare_device_data(device_info)
        self._update_network_function_device_db(device,
                                                'HEALTH_CHECK_COMPLETED')
        orchestration_driver = self._get_orchestration_driver(
            device['service_details']['service_vendor'])

        _ifaces_plugged_in = (
            orchestration_driver.plug_network_function_device_interfaces(
                device))
        if _ifaces_plugged_in:
            self._increment_device_interface_count(device)
            self._create_event(event_id='CONFIGURE_DEVICE',
                               event_data=device,
                               is_internal_event=True)
        else:
            self._create_event(event_id='DEVICE_CONFIGURATION_FAILED',
                               event_data=device,
                               is_internal_event=True)

    def plug_interfaces_fast(self, event):

        # In this case, the event will be
        # happening in parallel with HEALTHMONITORING,
        # so, we should not generate CONFIGURE_DEVICE & should not update
        # DB with HEALTH_CHECK_COMPLETED.

        nfp_context = event.context

        service_details = nfp_context['service_details']
        network_function_device = nfp_context['network_function_device']
        nf_id = network_function_device['id']
        nfi_id = nfp_context['network_function_instance']['id']
        token = nfp_context['resource_owner_context']['admin_token']
        tenant_id = nfp_context['resource_owner_context']['admin_tenant_id']

        consumer = nfp_context['consumer']
        provider = nfp_context['provider']

        event.key = nf_id + nfi_id

        orchestration_driver = self._get_orchestration_driver(
            service_details['service_vendor'])
        ports = self._make_ports_dict(
            nfp_context.get('explicit_consumer', consumer),
            provider, 'port')

        device = {
            'id': network_function_device['id'],
            'ports': ports,
            'service_details': service_details,
            'token': token,
            'tenant_id': tenant_id,
            'interfaces_in_use': network_function_device['interfaces_in_use'],
            'status': network_function_device['status'],
            'provider_metadata': nfp_context['provider_metadata'],
            'enable_port_security': nfp_context.get('enable_port_security')
        }

        _ifaces_plugged_in = (
            orchestration_driver.plug_network_function_device_interfaces(
                device))
        if _ifaces_plugged_in:
            # self._increment_device_interface_count(device)
            # [REVISIT(mak)] - Check how incremented ref count can be
            # updated in DB
            self._controller.event_complete(event, result="SUCCESS")
        else:
            self._create_event(event_id="PLUG_INTERFACE_FAILED",
                               is_internal_event=True)
            self._controller.event_complete(event, result="FAILED")

    def configure_device(self, event):
        device = event.data
        orchestration_driver = self._get_orchestration_driver(
            device['service_details']['service_vendor'])
        config_params = (
            orchestration_driver.get_network_function_device_config(
                device, nfp_constants.GENERIC_CONFIG))
        if not config_params:
            self._create_event(event_id='DRIVER_ERROR',
                               event_data=device,
                               is_internal_event=True)
            return None
        # Sends RPC to configurator to create generic config
        self.configurator_rpc.create_network_function_device_config(
            device, config_params)

    def create_device_configuration(self, event):
        nfp_context = event.context

        service_details = nfp_context['service_details']
        consumer = nfp_context['consumer']
        provider = nfp_context['provider']
        management = nfp_context['management']
        network_function = nfp_context['network_function']
        network_function_instance = nfp_context['network_function_instance']
        network_function_device = nfp_context['network_function_device']
        tenant_id = nfp_context['resource_owner_context']['admin_tenant_id']

        binding_key = service_details[
            'service_vendor'].lower() + network_function['id']

        orchestration_driver = self._get_orchestration_driver(
            service_details['service_vendor'])
        device = {
            'tenant_id': tenant_id,
            'mgmt_ip_address': management['port']['ip_address'],
            'mgmt_ip': network_function_device['mgmt_ip_address'],
            'provider_ip': provider['port']['ip_address'],
            'provider_cidr': provider['subnet']['cidr'],
            'provider_mac': provider['port']['mac_address'],
            'provider_gateway_ip': provider['subnet']['gateway_ip']}

        if consumer['port'] and consumer['subnet']:
            device.update({'consumer_ip': consumer['port']['ip_address'],
                           'consumer_cidr': consumer['subnet']['cidr'],
                           'consumer_mac': consumer['port']['mac_address'],
                           'consumer_gateway_ip': consumer[
                               'subnet']['gateway_ip']})

        nfp_context['event_desc'] = event.desc.to_dict()
        device.update({
            'id': network_function_device['id'],
            'mgmt_ip_address': network_function_device['mgmt_ip_address'],
            'service_details': service_details,
            'network_function_id': network_function['id'],
            'network_function_instance_id': network_function_instance['id'],
            'nfp_context': {
                'event_desc': nfp_context['event_desc'],
                'id': event.id, 'key': event.key,
                'network_function_device': network_function_device,
                'binding_key': binding_key}})

        config_params = (
            orchestration_driver.
            get_network_function_device_config(
                device, nfp_constants.GENERIC_CONFIG))

        if not config_params:
            self._create_event(event_id='DRIVER_ERROR',
                               event_data=device,
                               is_internal_event=True)
            self._controller.event_complete(event, result="FAILED")
            return None

        event_data = {'device': device, 'nfp_context': nfp_context,
                      'config_params': config_params}
        self._create_event(event_id='UPDATE_DEVICE_CONFIG_PARAMETERS',
                           event_data=event_data)

    def device_configuration_updated(self, event):
        nfp_context, config_params, device = (
            event.data['nfp_context'], event.data['config_params'],
            event.data['device'])
        # Set forward_route as False in resource_data for configurator to
        # handle routes differently, when vpn is in service chain
        if nfp_utils.is_vpn_in_service_chain(
                nfp_context['service_chain_specs']):
            for cfg in config_params['config']:
                cfg['resource_data']['forward_route'] = False
        else:
            for cfg in config_params['config']:
                cfg['resource_data']['forward_route'] = True
        # Sends RPC to configurator to create generic config
        self.configurator_rpc.create_network_function_device_config(
            device, config_params)
        self._controller.event_complete(event=event, result='SUCCESS')

    def configuration_complete(self, event):
        nfp_context = event.context
        nf_id = nfp_context['network_function']['id']
        event_results = event.result
        for result in event_results:
            if result.result.lower() != "success":
                device = self._prepare_failure_case_device_data(nfp_context)
                self._create_event(event_id='DEVICE_CREATE_FAILED',
                                   event_data=device)
                return self._controller.event_complete(event, result="FAILED")
        sc_event = self._controller.new_event(id="SERVICE_CONFIGURED",
                                              key=nf_id)
        self._controller.post_event(sc_event)
        self._controller.event_complete(event, result="SUCCESS")

    def device_configuration_complete(self, event, result='SUCCESS'):
        nfp_context = event.data['nfp_context']

        # Invoke event_complete for original event which is
        # CREATE_DEVICE_CONFIGURATION
        event_desc = nfp_context.pop('event_desc', None)
        key = nfp_context.pop('key', None)
        self._controller.event_complete(event)
        event = self._controller.new_event(id="CREATE_DEVICE_CONFIGURATION",
                                           key=key, desc_dict=event_desc)
        event.binding_key = nfp_context.pop('binding_key', None)
        self._controller.event_complete(event, result=result)

    def delete_network_function_device(self, event):
        network_function_details = event.context
        nfd = network_function_details['network_function_device']
        if not nfd:
            self._controller.event_complete(event, result="SUCCESS")
            return
        device = self._prepare_device_data_fast(network_function_details)
        LOG.info("Recieved DELETE NETWORK FUNCTION "
                 "DEVICE request ")
        device['event_desc'] = event.desc.to_dict()
        self._create_event(event_id='DELETE_CONFIGURATION',
                           event_data=device,
                           is_internal_event=True)

    def delete_device_configuration(self, event):
        device = event.data
        orchestration_driver = self._get_orchestration_driver(
            device['service_details']['service_vendor'])
        nfp_context = module_context.get()
        config_params = (
            orchestration_driver.get_network_function_device_config(
                device, nfp_constants.GENERIC_CONFIG, is_delete=True,
                devices_data=nfp_context))
        if not config_params:
            self._create_event(event_id='DRIVER_ERROR',
                               event_data=device,
                               is_internal_event=True)
            nf_id = device['network_function_id']
            dnfd_event = (
                self._controller.new_event(
                    id='DELETE_NETWORK_FUNCTION_DEVICE',
                    key=nf_id,
                    binding_key=nf_id,
                    desc_dict=device.get('event_desc')))
            self._controller.event_complete(dnfd_event, result='FAILED')
            # TODO(mak): If driver returns ERROR,
            # then we are not proceeding further
            # Stale vms will exist in this case.
            # Need to handle this case where
            # driver returned None So dont initiate configurator API but call
            # unplug_interfaces and device delete to delete vms.
            return None

        nf_data = {
            'service_chain_instance': nfp_context.get(
                'service_chain_instance'),
            'provider': nfp_context.get('provider'),
            'consumer': nfp_context.get('consumer')
        }
        device.update(nf_data)
        # Sends RPC call to configurator to delete generic config API
        self.configurator_rpc.delete_network_function_device_config(
            device, config_params)

    def unplug_interfaces(self, event):
        result = "SUCCESS"
        device = event.data
        self._decrement_device_ref_count(device)
        orchestration_driver = self._get_orchestration_driver(
            device['service_details']['service_vendor'])

        is_interface_unplugged = (
            orchestration_driver.unplug_network_function_device_interfaces(
                device))
        if is_interface_unplugged:
            mgmt_port_id = device['mgmt_port_id']
            self._decrement_device_interface_count(device)
            device['mgmt_port_id'] = mgmt_port_id
        else:
            result = "FAILED"
        self._create_event(event_id='DELETE_DEVICE',
                           event_data=device,
                           is_internal_event=True)
        self._controller.event_complete(event, result=result)

    def delete_configuration_complete(self, event):
        device = event.data['nfp_context']
        nfd_id = event.data['network_function_device_id']
        nf_id = event.data['network_function_id']
        unplug_interfaces = (
            self._controller.new_event(id='UNPLUG_INTERFACES',
                                       data=device,
                                       key=nf_id,
                                       binding_key=nfd_id,
                                       serialize=True))
        self._controller.post_event(unplug_interfaces)
        self._controller.event_complete(event)

    def delete_device(self, event):
        # Update status in DB, send DEVICE_DELETED event to NSO.
        device = event.data
        orchestration_driver = self._get_orchestration_driver(
            device['service_details']['service_vendor'])

        network_function = (
            self.nsf_db.get_network_function(
                self.db_session,
                device['network_function_id']))
        device['network_function'] = network_function
        chm_event = self._controller.new_event(
            id='PERFORM_CLEAR_HM',
            key=device['network_function_id'],
            data=device)
        self._controller.post_event(chm_event)

        orchestration_driver.delete_network_function_device(device)
        self._create_event(
            event_id='DEVICE_BEING_DELETED',
            event_data=device,
            is_poll_event=True,
            original_event=event,
            max_times=nfp_constants.DEVICE_BEING_DELETED_MAXRETRY)

    @nfp_api.poll_event_desc(
        event='DEVICE_BEING_DELETED',
        spacing=nfp_constants.DEVICE_BEING_DELETED_SPACING)
    def check_device_deleted(self, event):
        device = event.data
        orchestration_driver = self._get_orchestration_driver(
            device['service_details']['service_vendor'])
        status = orchestration_driver.get_network_function_device_status(
            device, ignore_failure=True)
        if not status:
            try:
                device_id = device['id']
                del device['id']
                orchestration_driver.delete_network_function_device(device)
                self._delete_network_function_device_db(device_id, device)
                if device.get('event_desc'):
                    nf_id = device['network_function_id']
                    dnfd_event = (
                        self._controller.new_event(
                            id='DELETE_NETWORK_FUNCTION_DEVICE',
                            key=nf_id,
                            binding_key=nf_id,
                            desc_dict=device['event_desc']))
                    self._controller.event_complete(
                        dnfd_event, result='SUCCESS')
                return STOP_POLLING
            except Exception as exc:
                device['id'] = device_id
                err = ("Exception - %s - in DEVICE_BEING_DELETED" % (exc))
                LOG.error(err)
                return CONTINUE_POLLING
        else:
            return CONTINUE_POLLING

    # Error Handling
    def handle_device_create_error(self, event):
        device = event.data
        LOG.error("Device creation failed, for device %(device)s",
                  {'device': device})
        device['network_function_device_id'] = device.get('id')
        self._create_event(event_id='DEVICE_CREATE_FAILED',
                           event_data=device)

    def handle_device_not_up(self, event):
        device = event.data
        self._release_cnfd_lock(device)
        status = nfp_constants.ERROR
        desc = 'Device not became ACTIVE'
        self._update_network_function_device_db(device, status, desc)
        device['network_function_device_id'] = device['id']
        self._create_event(event_id='DEVICE_CREATE_FAILED',
                           event_data=device)

    def _prepare_failure_case_device_data(self, nfp_context):
        network_function = nfp_context['network_function']
        network_function_instance = nfp_context['network_function_instance']
        device = {'network_function_id': network_function['id'],
                  'network_function_instance_id': network_function_instance[
            'id'], 'binding_key': nfp_context.get('binding_key')}
        network_function_device = nfp_context.get('network_function_device')
        if network_function_device:
            device.update(
                {'network_function_device_id': network_function_device['id']})
            device.update(network_function_device)
        return device

    def handle_plug_interface_failed(self, event):
        nfp_context = event.context
        device = self._prepare_failure_case_device_data(nfp_context)
        # self._release_cnfd_lock(device)
        status = nfp_context['network_function_device']['status']
        desc = "Failed to plug interfaces"
        self._update_network_function_device_db(device, status, desc)
        # self._create_event(event_id='DEVICE_CREATE_FAILED',
        #                    event_data=device)

    def handle_device_not_reachable(self, event):
        device = event.data
        status = nfp_constants.ERROR
        desc = 'Device not reachable, Health Check Failed'
        self._update_network_function_device_db(device, status, desc)
        device['network_function_device_id'] = device['id']
        # self._create_event(event_id='DEVICE_CREATE_FAILED',
        #                    event_data=device)
        self.health_monitor_complete(event, result='FAILED')

    def periodic_hm_handle_device_reachable(self, event):
        device = event.data
        status = nfp_constants.ACTIVE
        desc = 'Device is ACTIVE'
        self._update_network_function_device_db(device, status, desc)

    def periodic_hm_handle_device_not_reachable(self, event):
        device = event.data
        status = nfp_constants.ERROR
        desc = 'Device not reachable, Health Check Failed'
        self._update_network_function_device_db(device, status, desc)

    def handle_device_config_failed(self, event):
        # device = event.data
        nfp_context = event.data['nfp_context']

        device = nfp_context['network_function_device']
        status = device['status']
        desc = 'Configuring Device Failed.'
        self._update_network_function_device_db(device, status, desc)
        device['network_function_device_id'] = device['id']
        # self._create_event(event_id='DEVICE_CREATE_FAILED',
        #                    event_data=event.data)
        LOG.debug("Device create failed for device: %s, with "
                  "data: %s", (device['id'], device))
        self.device_configuration_complete(event, result='FAILED')

    def handle_interfaces_setup_failed(self, event):
        device = event.data
        status = nfp_constants.ERROR
        desc = 'Interfaces Plugging failed'
        self._update_network_function_device_db(device, status, desc)
        device['network_function_device_id'] = device['id']
        self._create_event(event_id='DEVICE_CREATE_FAILED',
                           event_data=device)
        LOG.debug("Interface Plugging failed for device: %s,"
                  "with config: %s", (device['id'], device))

    def handle_driver_error(self, event):
        device = event.data
        LOG.error("Exception occured in driver, driver returned None "
                  " for device %(device)s", {'device': device})
        status = nfp_constants.ERROR
        desc = 'Exception in driver, driver return None'
        self._update_network_function_device_db(device, status, desc)
        # device['network_function_device_id'] = device['id']
        # self._create_event(event_id='DEVICE_CREATE_FAILED',
        #                    event_data=device)

    def update_config_params(self, event):
        self._create_event(event_id='DEVICE_CONFIG_PARAMETERS_UPDATED',
                           event_data=event.data, is_internal_event=True)
        self._controller.event_complete(event=event, result='SUCCESS')


class NDOConfiguratorRpcApi(object):
    """Service Manager side of the Service Manager to Service agent RPC API"""
    API_VERSION = '1.0'
    target = messaging.Target(version=API_VERSION)

    def __init__(self, context, conf):
        super(NDOConfiguratorRpcApi, self).__init__()
        self.conf = conf
        self.context = context
        self.client = n_rpc.get_client(self.target)
        self.rpc_api = self.client.prepare(
            version=self.API_VERSION,
            topic=nsf_topics.NFP_NDO_CONFIGURATOR_TOPIC)

    def _get_request_info(self, device, operation):
        nfp_context = module_context.get()
        request_info = {
            'nf_id': device['network_function_id'],
            'nfi_id': (
                device['network_function_instance_id']),
            'nfd_id': device['id'],
            'requester': nfp_constants.DEVICE_ORCHESTRATOR,
            'operation': operation,
            'logging_context': nfp_context['log_context'],
            # So that notification callbacks can work on cached data
            # 'orig_nfp_context': device.get('orig_nfp_context'),
            'nfp_context': device.get('nfp_context', None),
            'service_profile': device.get('service_profile'),
            'service_vm_context': nfp_utils.get_service_vm_context(
                device['service_details']['service_vendor']),
        }
        nfd_ip = device.get('mgmt_ip_address')
        request_info.update({'device_ip': nfd_ip})
        return request_info

    def _update_params(self, device_data, config_params, operation):
        request_info = self._get_request_info(device_data, operation)
        if not config_params:
            return None
        config_params['info'] = {
            'service_type': device_data['service_details']['service_type'],
            'service_vendor': device_data['service_details']['service_vendor'],
            'context': request_info,
        }
        if device_data.get('service_feature'):
            config_params['info'].update(
                {'service_feature': device_data.get('service_feature')})
        if config_params.get('service_info'):
            config_params['info'].update(config_params.pop('service_info'))

    def create_network_function_device_config(self, device_data,
                                              config_params):
        self._update_params(device_data, config_params, operation='create')
        LOG.info("Sending create NFD config request to configurator "
                 "for NF:%(nf_id)s ",
                 {'nf_id': config_params['info']['context']['nf_id']})

        transport.send_request_to_configurator(self.conf,
                                               self.context,
                                               config_params,
                                               'CREATE',
                                               True)

    def delete_network_function_device_config(self, device_data,
                                              config_params):
        self._update_params(device_data, config_params, operation='delete')
        config_params['info']['context']['nfp_context'] = device_data
        LOG.info("Sending delete NFD config request to configurator ")

        transport.send_request_to_configurator(self.conf,
                                               self.context,
                                               config_params,
                                               'DELETE',
                                               True)


class ExceptionHandler(object):

    @staticmethod
    def event_method_mapping(event_id):
        event_handler_mapping = {
            "CREATE_NETWORK_FUNCTION_DEVICE": (
                ExceptionHandler.create_network_function_device),
            "DEVICE_SPAWNING": ExceptionHandler.device_spawning,
            "PERFORM_INITIAL_HEALTH_CHECK":
                ExceptionHandler.perform_initial_health_check,
            "DEVICE_UP": ExceptionHandler.device_up,
            "PLUG_INTERFACES": ExceptionHandler.plug_interfaces,
            "HEALTH_MONITOR_COMPLETE":
                ExceptionHandler.health_monitor_complete,
            "CREATE_DEVICE_CONFIGURATION":
                ExceptionHandler.create_device_configuration,
            "DEVICE_CONFIGURED":
                ExceptionHandler.device_configuration_complete,
            "CONFIGURATION_COMPLETE": ExceptionHandler.configuration_complete,
            "DELETE_NETWORK_FUNCTION_DEVICE": (
                ExceptionHandler.delete_network_function_device),
            "DELETE_CONFIGURATION":
                ExceptionHandler.delete_device_configuration,
            "DELETE_CONFIGURATION_COMPLETED": (
                ExceptionHandler.delete_configuration_complete),
            "UNPLUG_INTERFACES": ExceptionHandler.unplug_interfaces,
            "DELETE_DEVICE": ExceptionHandler.delete_device,
            "DEVICE_BEING_DELETED": ExceptionHandler.device_being_deleted,
            "PERIODIC_HM_DEVICE_NOT_REACHABLE": (
                ExceptionHandler.periodic_hm_handle_device_not_reachable),
            "DEVICE_NOT_REACHABLE": (
                ExceptionHandler.health_monitor_complete),
            "DEVICE_CONFIGURATION_FAILED": (
                ExceptionHandler.device_configuration_complete),
            "PERFORM_PERIODIC_HEALTH_CHECK": (
                ExceptionHandler.perform_periodic_health_check),
        }
        if event_id not in event_handler_mapping:
            raise Exception(_("Invalid event ID"))
        else:
            return event_handler_mapping[event_id]

    @staticmethod
    def handle(orchestrator, event, exception):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        message = "Traceback: %s" % traceback.format_exception(
            exc_type, exc_value, exc_traceback)
        LOG.error(message)

        exception_handler = ExceptionHandler.event_method_mapping(event.id)
        return exception_handler(orchestrator, event, exception)

    @staticmethod
    def create_network_function_device(orchestrator, event, exception):
        nfp_context = event.context
        network_function = nfp_context['network_function']
        # [REVISIT: AKASH] Updating NF from device_orchestrator is wrong way
        # of doing, but still doing it, will correct it later
        orchestrator.nsf_db.update_network_function(
            orchestrator.db_session,
            network_function['id'],
            {'status': nfp_constants.ERROR})
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def perform_initial_health_check(orchestrator, event, exception):
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def device_up(orchestrator, event, exception):
        nfp_context = event.context
        network_function = nfp_context['network_function']
        device = orchestrator._prepare_failure_case_device_data(nfp_context)
        orchestrator._release_cnfd_lock(device)
        orchestrator.nsf_db.update_network_function(
            orchestrator.db_session,
            network_function['id'],
            {'status': nfp_constants.ERROR})
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def plug_interfaces(orchestrator, event, exception):
        nfp_context = event.context
        device = orchestrator._prepare_failure_case_device_data(nfp_context)
        status = nfp_context['network_function_device']['status']
        desc = "Failed to plug interfaces"
        orchestrator._update_network_function_device_db(device, status, desc)
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def health_monitor_complete(orchestrator, event, exception):
        nfp_context = event.data['nfp_context']
        event_desc = nfp_context.pop('event_desc', None)
        nfp_context.pop('id', None)
        key = nfp_context.pop('key', None)
        ev = orchestrator._controller.new_event(
            id="PERFORM_INITIAL_HEALTH_CHECK",
            key=key, desc_dict=event_desc)
        orchestrator._controller.event_complete(ev, result='FAILED')
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def create_device_configuration(orchestrator, event, exception):
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def device_configuration_complete(orchestrator, event, exception):
        nfp_context = event.data['nfp_context']
        event_desc = nfp_context.pop('event_desc')
        key = nfp_context.pop('key')
        ev = orchestrator._controller.new_event(
            id="CREATE_DEVICE_CONFIGURATION",
            key=key, desc_dict=event_desc)
        ev.binding_key = nfp_context.pop('binding_key')
        orchestrator._controller.event_complete(ev, result='FAILED')
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def configuration_complete(orchestrator, event, exception):
        nfp_context = event.context
        network_function = nfp_context['network_function']
        orchestrator.nsf_db.update_network_function(
            orchestrator.db_session,
            network_function['id'],
            {'status': nfp_constants.ERROR})
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def delete_network_function_device(orchestrator, event, exception):
        network_function_details = event.context
        device = network_function_details['network_function_device']
        status = nfp_constants.ERROR
        desc = 'Exception in driver, driver return None'
        orchestrator._update_network_function_device_db(device, status, desc)
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def delete_device_configuration(orchestrator, event, exception):
        device = event.data
        status = nfp_constants.ERROR
        desc = 'Exception in driver, driver return None'
        orchestrator._update_network_function_device_db(device, status, desc)
        nf_id = device['network_function_id']
        dnfd_event = (
            orchestrator._controller.new_event(
                id='DELETE_NETWORK_FUNCTION_DEVICE',
                key=nf_id,
                binding_key=nf_id,
                desc_dict=device.get('event_desc')))
        orchestrator._controller.event_complete(dnfd_event, result='FAILED')
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def delete_configuration_complete(orchestrator, event, exception):
        device = event.data
        status = nfp_constants.ERROR
        desc = 'Exception in driver, driver return None'
        orchestrator._update_network_function_device_db(device, status, desc)
        nf_id = device['network_function_id']
        dnfd_event = (
            orchestrator._controller.new_event(
                id='DELETE_NETWORK_FUNCTION_DEVICE',
                key=nf_id,
                binding_key=nf_id,
                desc_dict=device.get('event_desc')))
        orchestrator._controller.event_complete(dnfd_event, result='FAILED')
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def unplug_interfaces(orchestrator, event, exception):
        device = event.data
        nf_id = device['network_function_id']
        dnfd_event = (
            orchestrator._controller.new_event(
                id='DELETE_NETWORK_FUNCTION_DEVICE',
                key=nf_id,
                binding_key=nf_id,
                desc_dict=device.get('event_desc')))
        orchestrator._controller.event_complete(dnfd_event, result='FAILED')
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def delete_device(orchestrator, event, exception):
        device = event.data
        status = nfp_constants.ERROR
        desc = 'Exception in driver, driver return None'
        orchestrator._update_network_function_device_db(device, status, desc)
        nf_id = device['network_function_id']
        dnfd_event = (
            orchestrator._controller.new_event(
                id='DELETE_NETWORK_FUNCTION_DEVICE',
                key=nf_id,
                binding_key=nf_id,
                desc_dict=device.get('event_desc')))
        orchestrator._controller.event_complete(dnfd_event, result='FAILED')
        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def device_being_deleted(orchestrator, event, exception):
        return {'poll': True}

    @staticmethod
    def device_spawning(orchestrator, event, exception):
        nfp_context = event.context
        network_function = nfp_context['network_function']
        device = orchestrator._prepare_failure_case_device_data(nfp_context)
        status = nfp_constants.ERROR
        desc = 'Exception in driver, driver return None'
        orchestrator._update_network_function_device_db(device, status, desc)
        orchestrator._release_cnfd_lock(device)
        orchestrator.nsf_db.update_network_function(
            orchestrator.db_session,
            network_function['id'],
            {'status': nfp_constants.ERROR})
        orchestrator._controller.event_complete(event, result='FAILED')
        return {'poll': False}

    @staticmethod
    def periodic_hm_handle_device_not_reachable(orchestrator,
                                                event, exception):

        orchestrator._controller.event_complete(event, result='FAILED')

    @staticmethod
    def perform_periodic_health_check(orchestrator, event, exception):
        orchestrator._controller.event_complete(event, result='FAILED')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():
    # Run the Testcases:
    test = test_gbp_prs_func()
    if test.test_gbp_prs_func_1() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PRS_FUNC_1')
    if test.test_gbp_prs_func_2() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PRS_FUNC_2')
    if test.test_gbp_prs_func_3() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PRS_FUNC_3')
    if test.test_gbp_prs_func_4() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PRS_FUNC_4')
    if sys.argv[1] == 'aci':
        test._log.info(
            "\nTESTCASE_GBP_PRS_FUNC_5: TO CREATE/VERIFY/DELETE/VERIFY "
            "a PARENT and CHILD POLICY RULESET\n")
        test._log.info("\nTESTCASE_GBP_PRS_FUNC_5: NOT SUPPORTED in ACI")
        test._log.info(
            "\nTESTCASE_GBP_PRS_FUNC_6: TO CHANGE/UPDATE/DELETE/VERIFY "
            "PARENT and CHILD POLICY RULESET\n")
        test._log.info("\nTESTCASE_GBP_PRS_FUNC_6: NOT SUPPORTED in ACI")
    else:
        if test.test_gbp_prs_func_5() == 0:
            test.cleanup(tc_name='TESTCASE_GBP_PRS_FUNC_5')
        if test.test_gbp_prs_func_6() == 0:
            test.cleanup(tc_name='TESTCASE_GBP_PRS_FUNC_6')
    test.cleanup()
    utils_libs.report_results('test_gbp_prs_func', 'test_results.txt')
    sys.exit(1)


class test_gbp_prs_func(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_prs_func.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_prs_func.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info(
            "\n## START OF GBP POLICY_RULE_SET FUNCTIONALITY TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.act_name = 'demo_pa'
        self.cls_name = 'demo_pc'
        self.rule_name = 'demo_pr'
        self.ruleset_name = 'demo_prs'
        self._log.info('\n## Step 1: Create a PC needed for PRS Testing ##')
        self.cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', self.cls_name)
        if self.cls_uuid == 0:
            self._log.info(
                "\nReqd Policy Classifier Create Failed, hence GBP "
                "Policy Rule-Set Functional Test Suite Run ABORTED\n")
            return
        self._log.info('\n## Step 1: Create a PA needed for PRS Testing ##')
        self.act_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'action', self.act_name)
        if self.act_uuid == 0:
            self._log.info(
                "\nReqd Policy Action Create Failed, hence GBP "
                "Policy Rule-Set Functional Test Suite Run ABORTED\n")
            return
        self._log.info('\n## Step 1: Create a PR needed for PRS Testing ##')
        self.rule_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'rule', self.rule_name, classifier=self.cls_name,
            action=self.act_name)
        if self.rule_uuid == 0:
            self._log.info(
                "\nReqd Policy Rule Create Failed, hence GBP "
                "Policy Rule-Set Functional Test Suite Run ABORTED\n ")
            return

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('%s: FAILED' % (tc_name))
        for obj in ['ruleset', 'rule', 'classifier', 'action']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_gbp_prs_func_1(
            self,
            name_uuid='',
            ruleset_uuid='',
            rep_cr=0,
            rep_del=0):

        if rep_cr == 0 and rep_del == 0:
            self._log.info(
                "\n########################################################\n"
                "TESTCASE_GBP_PRS_FUNC_1: TO CREATE/VERIFY/DELETE/VERIFY "
                "a POLICY RULESET with DEFAULT ATTRIB VALUE\n"
                "TEST_STEP::\n"
                "Create Policy RuleSet Object\n"
                "Verify the attributes & value, show & list cmds\n"
                "Delete Policy RuleSet using Name\n"
                "Verify that PR has got deleted, show & list cmds\n"
                "##########################################################\n")

        if name_uuid == '':
            name_uuid = self.ruleset_name
        # Testcase work-flow starts
        if rep_cr == 0 or rep_cr == 1:
            self._log.info(
                '\n## Step 1: Create RuleSet with default attrib vals##\n')
            ruleset_uuid = self.gbpcfg.gbp_policy_cfg_all(
                1, 'ruleset', name_uuid)
            if ruleset_uuid == 0:
                self._log.info("# Step 1: Create RuleSet == Failed")
                return 0
            self._log.info('# Step 2A: Verify RuleSet using -list cmd')
            if self.gbpverify.gbp_policy_verify_all(
                    0, 'ruleset', name_uuid, ruleset_uuid) == 0:
                self._log.info(
                    "# Step 2A: Verify RuleSet using -list option == Failed")
                return 0
            self._log.info('# Step 2B: Verify RuleSet using -show cmd')
            if self.gbpverify.gbp_policy_verify_all(
                    1, 'ruleset', name_uuid, id=ruleset_uuid,
                    shared='False') == 0:
                self._log.info(
                    "# Step 2B: Verify RuleSet using -show option == Failed")
                return 0
        #######
        if rep_del == 0 or rep_del == 1:
            self._log.info('\n## Step 3: Delete RuleSet using name  ##\n')
            if self.gbpcfg.gbp_policy_cfg_all(0, 'ruleset', name_uuid) == 0:
                self._log.info("# Step 3: Delete RuleSet == Failed")
                return 0
            if self.gbpverify.gbp_policy_verify_all(
                    0, 'ruleset', name_uuid, ruleset_uuid) != 0:
                self._log.info(
                    "# Step 3A: Verify RuleSet is Deleted using -list "
                    "option == Failed")
                return 0
            if self.gbpverify.gbp_policy_verify_all(
                    1, 'ruleset', name_uuid, id=ruleset_uuid,
                    shared='False') != 0:
                self._log.info(
                    "# Step 3B: Verify RuleSet is Deleted using "
                    "-show option == Failed")
                return 0
            if rep_cr == 0 and rep_del == 0:
                self._log.info("\nTESTCASE_GBP_PRS_FUNC_1: PASSED")
        return 1

    def test_gbp_prs_func_2(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PRS_FUNC_2: TO CREATE/VERIFY/DELETE/VERIFY "
            "a POLICY RULESET with POLICY RULE\n"
            "TEST_STEP::\n"
            "Create Policy RuleSet Object with GBP PR\n"
            "Verify the attributes & value, show & list cmds\n"
            "Delete Policy RuleSet using Name\n"
            "Verify that PR has got deleted, show & list cmds\n"
            "##############################################################\n")

        # Testcase work-flow starts
        self._log.info("\n## Step 1: Create Policy RuleSet with PR ##")
        ruleset_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'ruleset', self.ruleset_name, policy_rules=self.rule_name)
        if ruleset_uuid == 0:
            self._log.info("# Step 1: Create RuleSet == Failed")
            return 0
        self._log.info('# Step 2A: Verify RuleSet using -list cmd')
        if self.gbpverify.gbp_policy_verify_all(
                0, 'ruleset', self.ruleset_name, ruleset_uuid) == 0:
            self._log.info(
                "# Step 2A: Verify RuleSet using -list option == Failed")
            return 0
        self._log.info('# Step 2B: Verify RuleSet using -show cmd')
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                self.ruleset_name,
                id=ruleset_uuid,
                policy_rules=self.rule_uuid,
                shared='False') == 0:
            self._log.info(
                "# Step 2B: Verify RuleSet using -show option == Failed")
            return 0
        self.test_gbp_prs_func_1(ruleset_uuid=ruleset_uuid, rep_cr=2)
        self._log.info("\nTESTCASE_GBP_PRS_FUNC_2: PASSED")
        return 1

    def test_gbp_prs_func_3(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PRS_FUNC_3: TO UPDATE/VERIFY/DELETE/VERIFY "
            "EACH ATTRIB of a POLICY RULESET\n"
            "TEST_STEP::\n"
            "Create Policy RuleSet using Default param values\n"
            "Update Each the Policy Rule's editable params\n"
            "Verify the Policy Rule's attributes & values, show & list cmds\n"
            "Delete the Policy Rule\n"
            "Verify Policy RuleSet successfully deleted\n"
            "##############################################################\n")

        # Testcase work-flow starts
        self._log.info('\n## Step 1: Create Policy RuleSet with PR ##\n')
        ruleset_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'ruleset', self.ruleset_name, policy_rules=self.rule_name)
        if ruleset_uuid != 0:
            self._log.info(
                "Step 1: Create RuleSet Passed, UUID == %s\n" %
                (ruleset_uuid))
        else:
            self._log.info("# Step 1: Create RuleSet == Failed")
            return 0
        self._log.info('\n## Step 1A: Create new PA ,new PC, new PR##\n')
        new_cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', 'grppol_pc1')
        if new_cls_uuid == 0:
            self._log.info(
                "\nNew Classifier Create Failed, hence "
                "testcase_gbp_prs_func_3 ABORTED\n")
            return 0
        new_act_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'action', 'grppol_pa1')
        if new_act_uuid == 0:
            self._log.info(
                "\nNew Action Create Failed, hence "
                "testcase_gbp_prs_func_3 ABORTED\n")
            return 0
        new_rule_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'rule',
            'grppol_pr',
            classifier=new_cls_uuid,
            action=new_act_uuid,
            description="'For devstack demo'")
        if new_rule_uuid == 0:
            self._log.info(
                "\nNew Rule Create Failed, hence "
                "testcase_gbp_prs_func_3 ABORTED\n")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(
                2,
                'ruleset',
                ruleset_uuid,
                name='grppol_prs',
                policy_rule=new_rule_uuid,
                description="'For devstack demo'") == 0:
            self._log.info(
                "\nStep 2: Updating Policy RuleSet's Attributes , Failed")
            return 0
        # Verify starts
        if self.gbpverify.gbp_policy_verify_all(
                0, 'ruleset', 'grppol_prs', ruleset_uuid) == 0:
            self._log.info(
                "# Step 2A: Verify Policy RuleSet Updated "
                "Attributes using -list option == Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                ruleset_uuid,
                name='grppol_prs',
                policy_rules=new_rule_uuid,
                description='For devstack demo') == 0:
            self._log.info(
                "# Step 2B: Verify Policy RuleSet Updated "
                "Attributes using -show option == Failed")
            return 0
        self._log.info("\nTESTCASE_GBP_PRS_FUNC_3: PASSED")
        return 1

    def test_gbp_prs_func_4(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PRS_FUNC_4: TO CREATE/UPDATE/VERIFY/DELETE/ "
            "ASSOCIATING MULTIPLE PRs to 1 POLICY RULESET \n"
            "TEST_STEP::\n"
            "Create Multiple Policy Rules\n"
            "Create Policy RuleSet by associating all the Policy Rules\n"
            "Verify that multiple Policy Rules are associated to the "
            "Policy RuleSet\n"
            "Update the Policy RuleSet such that few Policy Rules "
            "are unmapped\n"
            "Verify the Policy Rule's attributes & values, show & list cmds\n"
            "Update the Policy RuleSet such that all Policy Rules "
            "association removed\n"
            "Verify the Policy Rule's attributes & values, show & list cmds\n"
            "##############################################################\n")

        # Testcase work-flow starts
        self._log.info(
            '\n## Step 1A: Create new PA ,new PC, 4 PRs using the '
            'same PA & PC##\n')
        new_cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', 'grppol_pc1')
        if new_cls_uuid == 0:
            self._log.info(
                "\nNew Classifier Create Failed, hence "
                "testcase_gbp_prs_func_4 ABORTED\n")
            return 0
        new_act_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'action', 'grppol_pa1')
        if new_act_uuid == 0:
            self._log.info(
                "\nNew Action Create Failed, hence "
                "testcase_gbp_prs_func_4 ABORTED\n")
            return 0
        rule_uuid_list = []
        for i in range(4):
            new_rule_uuid = self.gbpcfg.gbp_policy_cfg_all(
                1,
                'rule',
                'grppol_pr_%s' %
                (i),
                classifier=new_cls_uuid,
                action=new_act_uuid,
                description="'For devstack demo'")
            if new_rule_uuid == 0:
                self._log.info(
                    "\nNew Rule Create Failed, hence "
                    "testcase_gbp_prs_func_4 ABORTED\n")
                return 0
            rule_uuid_list.append(new_rule_uuid)
        ruleset_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'ruleset',
            'grppol_prs_many',
            policy_rule='"%s %s %s %s"' %
            (rule_uuid_list[0],
             rule_uuid_list[1],
             rule_uuid_list[2],
             rule_uuid_list[3]),
            description="'For devstack demo'")
        if ruleset_uuid == 0:
            self._log.info(
                "\nStep 2: Updating Policy RuleSet's Attributes , Failed")
            return 0
        # Verify starts
        if self.gbpverify.gbp_policy_verify_all(
                0, 'ruleset', 'grppol_prs_many', ruleset_uuid) == 0:
            self._log.info(
                "# Step 2A: Verify Policy RuleSet Updated Attributes "
                "using -list option == Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                ruleset_uuid,
                name='grppol_prs_many',
                description='For devstack demo') == 0:
            self._log.info(
                "# Step 2B: Verify Policy RuleSet Updated Attributes "
                "using -show option == Failed")
            return 0
        if self.gbpverify.gbp_obj_ver_attr_all_values(
                'ruleset', 'grppol_prs_many', 'policy_rules',
                rule_uuid_list) == 0:
            self._log.info(
                "# Step 2C: Verify Policy RuleSet and its Multiple PRs "
                "using -show option == Failed")
            return 0
        # Update and Verify the PRS by updating the PRs(removing few existing
        # ones)
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'ruleset', 'grppol_prs_many', policy_rule='"%s %s"' %
                (rule_uuid_list[0], rule_uuid_list[2])) == 0:
            self._log.info(
                "# Step 3: Updating Policy RuleSet's Attributes , Failed")
            return 0
        if self.gbpverify.gbp_obj_ver_attr_all_values(
                'ruleset', 'grppol_prs_many', 'policy_rules',
                rule_uuid_list) != 0:
            self._log.info(
                "# Step 3A: Verify Policy RuleSet and its Multiple "
                "PRs using -show option == Failed")
            return 0
        # Update and Verify the PRS by updating the PRs=NULL(unmapping all PRs)
        if self.gbpcfg.gbp_policy_cfg_all(2, 'ruleset', 'grppol_prs_many',
                                          policy_rule='""') == 0:
            self._log.info(
                "# Step 4: Upmapping All Policy Rule from Policy "
                "RuleSet , Failed")
            return 0
        if self.gbpverify.gbp_obj_ver_attr_all_values(
                'ruleset', 'grppol_prs_many', 'policy_rules',
                rule_uuid_list) != 0:
            self._log.info(
                "# Step 4A: Verify All Policy Rules have been Removed "
                "from Policy RuleSet using --show option == Failed")
            return 0
        self._log.info("\nTESTCASE_GBP_PRS_FUNC_4: PASSED")
        return 1

    def test_gbp_prs_func_5(self):
        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PRS_FUNC_5: TO CREATE/VERIFY/DELETE/VERIFY "
            "a PARENT and CHILD POLICY RULESET\n"
            "TEST_STEP::\n"
            "Create 1 Policy RuleSet using the same PA & PC\n"
            "Create the 2nd Policy RuleSet using the same PA, "
            "PC and associate PRS-1 as CHILD\n"
            "Verify the Child PRS reflect the Parent PRS and viceversa\n"
            "Delete the Child PRS\n"
            "Verify the Parent PRS has no CHILD\n"
            "Create the CHild PRS, associate to the Parent PRS\n"
            "Verify the association is established b/e Child and Parent\n"
            "Delete the Parent PRS\n"
            "Verify the Parent PRS association removed the Child PRS\n"
            "##############################################################\n")

        # Testcase work-flow starts
        self._log.info("\n## Step 1: Create Policy RuleSet with PR ##")
        child_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'ruleset', 'demo_child_prs', policy_rules=self.rule_name)
        if child_uuid == 0:
            self._log.info(
                "\n## Step 1: Create Child Policy RuleSet == Failed")
            return 0
        parent_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1,
            'ruleset',
            'demo_par_prs',
            policy_rules=self.rule_name,
            child_policy_rule_sets=child_uuid)
        if parent_uuid == 0:
            self._log.info(
                "\n## Step 2: Create Parent Policy RuleSet == Failed")
            return 0
        self._log.info(
            '# Step 2A: Verify Parent and Child Policy RuleSet '
            'using -show cmd')
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                'demo_par_prs',
                id=parent_uuid,
                policy_rules=self.rule_uuid,
                shared='False',
                child_policy_rule_sets=child_uuid) == 0:
            self._log.info(
                "\n## Step 2A: Verify Parent RuleSet using -show "
                "option == Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                'demo_child_prs',
                id=child_uuid,
                policy_rules=self.rule_uuid,
                shared='False',
                parent_id=parent_uuid) == 0:
            self._log.info(
                "\n## Step 2B: Verify Parent RuleSet using -show "
                "option == Failed")
            return 0
        # Delete Child PRS
        if self.gbpcfg.gbp_policy_cfg_all(0, 'ruleset', 'demo_child_prs') == 0:
            self._log.info("# Step 3: Delete Child Policy RuleSet == Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                'demo_par_prs',
                id=parent_uuid,
                policy_rules=self.rule_uuid,
                shared='False',
                child_policy_rule_sets=child_uuid) != 0:
            self._log.info(
                "# Step 3A: Verify Parent PRS after Delete of Child "
                "PRS using -show option == Failed")
            return 0
        # Create Child PRS,Associate to Parent and Verify
        child_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'ruleset', 'demo_child_prs', policy_rules=self.rule_name)
        if child_uuid == 0:
            self._log.info("# Step 4: Create Child Policy RuleSet == Failed")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(
                2,
                'ruleset',
                'demo_par_prs',
                child_policy_rule_sets=child_uuid) == 0:
            self._log.info(
                "# Step 5: Associating Child PRS by Updating Parent "
                "PRS == Failed")
            return 0
        self._log.info(
            '# Step 5A: Verify Parent and Child Policy RuleSet '
            'using -show cmd')
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                'demo_par_prs',
                id=parent_uuid,
                policy_rules=self.rule_uuid,
                shared='False',
                child_policy_rule_sets=child_uuid) == 0:
            self._log.info(
                "# Step 5A: Verify Parent RuleSet using -show "
                "option == Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                'demo_child_prs',
                id=child_uuid,
                policy_rules=self.rule_uuid,
                shared='False',
                parent_id=parent_uuid) == 0:
            self._log.info(
                "# Step 5B: Verify Parent RuleSet using -show "
                "option == Failed")
            return 0
        # Delete Parent PRS and Verify
        if self.gbpcfg.gbp_policy_cfg_all(0, 'ruleset', 'demo_par_prs') == 0:
            self._log.info("# Step 6: Delete Parent Policy RuleSet == Failed")
            return 0
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'ruleset',
                'demo_child_prs',
                id=child_uuid,
                policy_rules=self.rule_uuid,
                shared='False',
                parent_id=parent_uuid) != 0:
            self._log.info(
                "# Step 6A: Verify Child PRS after Delete of "
                "Parent PRS using -show option == Failed")
            return 0
        self._log.info("\nTESTCASE_GBP_PRS_FUNC_5: PASSED")
        return 1

    def test_gbp_prs_func_6(self):
        """
        Changing parent-child prs mapping
        Create 4 PRS, two are parent and two are child
        Update the one of the parent such that both childs are mapped
        Verify
        Delete both child PRS
        Verify the parent PRS
        """
        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_PRS_FUNC_6: TO CHANGE/UPDATE/DELETE/VERIFY "
            "PARENT and CHILD POLICY RULESET\n"
            "TEST_STEP::\n"
            "Changing parent-child prs mapping\n"
            "Create 4 PRS, two are parent and two are child\n"
            "Update the one of the parent such that both childs are mapped\n"
            "Verify\n"
            "Delete both child PRS\n"
            "Verify the parent PRS\n"
            "##############################################################\n")

        # Testcase work-flow starts
        self._log.info(
            "\n## Step 1: Create 4 Policy RuleSets, 2 Parent & 2 "
            "Child with PR ##")
        ch_uuids, par_uuids = [], []
        for i in range(1, 3):
            child_name, par_name = 'child_%s_prs' % (i), 'par_%s_prs' % (i)
            child_uuid = self.gbpcfg.gbp_policy_cfg_all(
                1, 'ruleset', child_name, policy_rules=self.rule_name)
            if child_uuid == 0:
                self._log.info(
                    "\n## Step 1: Create Child Policy RuleSet == Failed")
                return 0
            parent_uuid = self.gbpcfg.gbp_policy_cfg_all(
                1,
                'ruleset',
                par_name,
                policy_rules=self.rule_name,
                child_policy_rule_sets=child_uuid)
            if parent_uuid == 0:
                self._log.info(
                    "\n## Step 2: Create Parent Policy RuleSet == Failed")
                return 0
            ch_uuids.append(child_uuid)
            par_uuids.append(parent_uuid)
        # Update One of the Parent with two Child PRSs
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'ruleset', 'par_1_prs', child_policy_rule_sets='"%s %s"' %
                (ch_uuids[0], ch_uuids[1])) == 0:
            self._log.info(
                "\n## Step 3: Update Parent Policy RuleSet == Failed")
            return 0
        if self.gbpverify.gbp_obj_ver_attr_all_values(
            'ruleset', 'par_1_prs', 'child_policy_rule_sets', [
                ch_uuids[0], ch_uuids[1]]) == 0:
            self._log.info("\n## Step 3A: Child PRS NOT Found in Parent PRS")
            return 0
        # Delete Child PRSs
        for i in range(2):
            if self.gbpcfg.gbp_policy_cfg_all(0, 'ruleset', ch_uuids[i]) == 0:
                self._log.info(
                    "\n## Step 5: Delete of Child PRS child_%s_prs" %
                    (i))
                return 0
        if self.gbpverify.gbp_obj_ver_attr_all_values(
            'ruleset', 'par_1_prs', 'child_policy_rule_sets', [
                ch_uuids[0], ch_uuids[1]]) != 0:
            self._log.info(
                "\n## Step 5A: Stale Child PRS Mapping still "
                "persists in Parent PRS")
            return 0
        self._log.info("\nTESTCASE_GBP_PRS_FUNC_6: PASSED")
        return 1


if __name__ == '__main__':
    main()
